#!/usr/bin/env python3
import torch
import torch.nn as nn
import torch.nn.functional as F

# from torch.nn.utils import _stateless

from typing import Iterable, Union


def fim_diag(module: nn.Module, _input: torch.Tensor,
             parameters: Iterable[nn.Parameter] = None
             ) -> list[torch.Tensor]:
    if parameters is None:
        parameters = tuple(module.parameters())
    _output = module(_input)  # (N, C)
    with torch.no_grad():
        prob = F.softmax(_output, dim=1).unsqueeze(-1)  # (N, C, 1)
    log_prob = F.log_softmax(_output, dim=1)  # (N, C)
    fim_dict: dict[int, list[torch.Tensor]] = {i: [] for i in range(len(parameters))}
    N, C = log_prob.shape
    for n in range(N):
        for c in range(C):
            grad_list = torch.autograd.grad(log_prob[n][c], parameters, retain_graph=True)
            for i, grad in enumerate(grad_list):    # different layers
                fim = grad.flatten().square()    # (D)
                fim_dict[i].append(fim.detach().clone())
    fim_list: list[torch.Tensor] = []
    for i, value in fim_dict.items():    # different layers
        D = value[0].shape[0]
        fim = torch.stack(value).view(N, C, D) * prob    # (N, C, D)
        fim = fim.sum(1).mean(0)   # (D)
        fim_list.append(fim)
    return fim_list


def fim(module: nn.Module, _input: torch.Tensor,
        parameters: Iterable[nn.Parameter] = None
        ) -> list[torch.Tensor]:
    if parameters is None:
        parameters = tuple(module.parameters())
    _output = module(_input)  # (N, C)
    with torch.no_grad():
        prob = F.softmax(_output, dim=1).unsqueeze(-1).unsqueeze(-1)  # (N, C, 1, 1)
    log_prob = F.log_softmax(_output, dim=1)  # (N, C)
    fim_dict: dict[int, list[torch.Tensor]] = {i: [] for i in range(len(parameters))}
    N, C = log_prob.shape
    for n in range(N):
        for c in range(C):
            grad_list = torch.autograd.grad(log_prob[n][c], parameters, retain_graph=True)
            for i, grad in enumerate(grad_list):    # different layers
                flatten_grad = grad.flatten()    # (D)
                fim = flatten_grad.unsqueeze(1) * flatten_grad.unsqueeze(0)   # (D, D)
                fim_dict[i].append(fim.detach().clone())
    fim_list: list[torch.Tensor] = []
    for i, value in fim_dict.items():    # different layers
        D = value[0].shape[0]
        fim = torch.stack(value).view(N, C, D, D) * prob    # (N, C, D, D)
        fim_list.append(fim.sum(1).mean(0))   # (D, D)
    return fim_list


# def new_fim(module: nn.Module, _input: torch.Tensor,
#             parameters: Union[dict[str, nn.Parameter], Iterable[nn.Parameter]] = None
#             ) -> list[torch.Tensor]:
#     if not isinstance(parameters, dict):
#         id_map: dict[torch.Tensor, str] = {param.data: name
#                                            for name, param in module.named_parameters()}
#         parameters = {id_map[param.data if isinstance(param, torch.nn.Parameter) else param]: param
#                       for param in parameters}
#     if parameters is None:
#         parameters = dict(module.named_parameters())
#     with torch.no_grad():
#         _output = module(_input)  # (N, C)
#         prob = F.softmax(_output, dim=1).unsqueeze(-1).unsqueeze(-1)  # (N, C, 1, 1)
#     keys, values = zip(*parameters.items())

#     def func(*params: torch.Tensor):
#         _output = _stateless.functional_call(module, {n: p for n, p in zip(keys, params)}, _input)
#         return F.log_softmax(_output, dim=1)  # (N, C)
#     jacobian_list: tuple[torch.Tensor] = torch.autograd.functional.jacobian(func, values)

#     fim_list: list[torch.Tensor] = []
#     for jacobian in jacobian_list:  # TODO: parallel
#         jacobian = jacobian.flatten(start_dim=2)   # (N, C, D)
#         fim = prob * jacobian.unsqueeze(-1) * jacobian.unsqueeze(-2)  # (N, C, D, D)
#         fim_list.append(fim.sum(1).mean(0))   # (D, D)
#     return fim_list



#!/usr/bin/env python3

from autovul.base.models import Model
import torch
import torch.nn as nn
import numpy as np
from torch.utils.hooks import RemovableHandle
from collections.abc import Callable

__all__ = ['register_hook', 'remove_hook']


def backward_hook(gamma: float) -> Callable[[nn.Module, torch.Tensor, torch.Tensor], torch.Tensor]:
    # implement SGM through grad through ReLU
    def _backward_hook(module: nn.Module, grad_in: torch.Tensor, grad_out: torch.Tensor):
        if isinstance(module, nn.ReLU):
            return (gamma * grad_in[0],)
    return _backward_hook


# def backward_hook_norm(module: nn.Module, grad_in: torch.Tensor, grad_out: torch.Tensor):
#     # normalize the gradient to avoid gradient explosion or vanish
#     std = torch.std(grad_in[0])
#     return (grad_in[0] / std,)


def register_hook_for_resnet(model: Model, gamma: float = 1.0) -> list[RemovableHandle]:
    # There is only 1 ReLU in Conv module of ResNet-18/34
    # and 2 ReLU in Conv module ResNet-50/101/152
    if model.layer in [50, 101, 152]:
        gamma = np.power(gamma, 0.5)
    backward_hook_sgm = backward_hook(gamma)

    _list: list[RemovableHandle] = []
    for name, module in model.named_modules():
        if 'relu' in name and '0.relu' not in name:
            _list.append(module.register_backward_hook(backward_hook_sgm))
    return _list
    # e.g., 1.layer1.1, 1.layer4.2, ...
    # if len(name.split('.')) == 3:
    #     module.register_backward_hook(backward_hook_norm)


def register_hook_for_densenet(model: Model, gamma: float = 1.0):
    # There are 2 ReLU in Conv module of DenseNet-121/169/201.
    gamma = np.power(gamma, 0.5)
    backward_hook_sgm = backward_hook(gamma)

    _list: list[RemovableHandle] = []
    for name, module in model.named_modules():
        if 'relu' in name and 'transition' not in name:
            _list.append(module.register_backward_hook(backward_hook_sgm))
    return _list


def register_hook(model: Model, gamma: float = 1.0):
    if 'sgm_remove' in model.__dict__.keys():
        print('SGM is already activated when calling register_hook')
        return
    if 'resnet' in model.name:
        model.sgm_remove = register_hook_for_resnet(model, gamma)
    elif 'densenet' in model.name:
        model.sgm_remove = register_hook_for_densenet(model, gamma)
    else:
        raise ValueError(model.name)


def remove_hook(model: Model):
    if 'sgm_remove' not in model.__dict__.keys():
        print('SGM is not activated when calling remove_hook')
        return
    for handle in model.sgm_remove:
        handle: RemovableHandle
        handle.remove()
    del model.sgm_remove



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet

import torchvision.datasets as datasets


class CIFAR10(ImageSet):

    name = 'cifar10'
    num_classes = 10
    data_shape = [3, 32, 32]
    class_to_idx = {'airplane': 0, 'automobile': 1, 'bird': 2, 'cat': 3,
                    'deer': 4, 'dog': 5, 'frog': 6, 'horse': 7, 'ship': 8, 'truck': 9}

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.49139968, 0.48215827, 0.44653124],
                                                           'std': [0.24703233, 0.24348505, 0.26158768], },
                 **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize(self):
        datasets.CIFAR10(root=self.folder_path, train=True, download=True)
        datasets.CIFAR10(root=self.folder_path, train=False, download=True)

    def _get_org_dataset(self, mode: str, **kwargs) -> datasets.CIFAR10:
        assert mode in ['train', 'valid']
        return datasets.CIFAR10(root=self.folder_path, train=(mode == 'train'), **kwargs)


class CIFAR100(CIFAR10):
    name = 'cifar100'
    num_classes = 100

    def initialize(self):
        datasets.CIFAR100(root=self.folder_path, train=True, download=True)
        datasets.CIFAR100(root=self.folder_path, train=False, download=True)

    def _get_org_dataset(self, mode: str, **kwargs) -> datasets.CIFAR100:
        assert mode in ['train', 'valid']
        return datasets.CIFAR100(root=self.folder_path, train=(mode == 'train'), **kwargs)



#!/usr/bin/env python3

from collections import OrderedDict
from .operations import get_op
from .genotypes import Genotype

import torch
import torch.nn as nn


class Cell(nn.Module):
    def __init__(self, genotype: Genotype,
                 C_prev_prev: int, C_prev: int, C: int,
                 reduction: bool, reduction_prev: bool,
                 dropout_p: float = None, std_conv: bool = False):
        super().__init__()
        op_name = 'factorized_reduce' if reduction_prev else 'conv'
        self.preprocess0 = get_op(op_name, C_prev_prev, C_out=C, std_conv=std_conv)
        self.preprocess1 = get_op('conv', C_prev, C_out=C, std_conv=std_conv)

        if reduction:
            op_names, indices = zip(*genotype.reduce)
            concat = genotype.reduce_concat
        else:
            op_names, indices = zip(*genotype.normal)
            concat = genotype.normal_concat
        op_names: list[str] = list(op_names)
        indices: list[int] = list(indices)
        concat: list[int]
        self._steps = len(op_names) // 2
        self._indices = indices
        self._concat = concat
        self.multiplier = len(concat)

        self._ops = nn.ModuleList()
        for name, index in zip(op_names, indices):
            stride = 2 if reduction and index < 2 else 1
            op = get_op(name, C, stride, p=dropout_p, std_conv=std_conv)
            self._ops.append(op)

    def forward(self, s0: torch.Tensor, s1: torch.Tensor) -> torch.Tensor:
        s0 = self.preprocess0(s0)
        s1 = self.preprocess1(s1)

        states = [s0, s1]
        for i in range(self._steps):
            h1 = states[self._indices[2 * i]]
            h2 = states[self._indices[2 * i + 1]]
            op1 = self._ops[2 * i]
            op2 = self._ops[2 * i + 1]
            states.append(op1(h1) + op2(h2))
        return torch.cat([states[i] for i in self._concat], dim=1)


# stride = 3 if CIFAR10
# stride = 2 if ImageNet
def AuxiliaryHead(C: int, num_classes: int = 10, stride: int = 3) -> nn.Sequential:
    """assuming input size 8x8"""
    return nn.Sequential(OrderedDict([
        ('features', nn.Sequential(
            nn.ReLU(),
            nn.AvgPool2d(5, stride=stride, padding=0, count_include_pad=False),  # image size = 2 x 2
            nn.Conv2d(C, 128, 1, bias=False),
            nn.BatchNorm2d(128),
            nn.ReLU(inplace=True),
            nn.Conv2d(128, 768, 2, bias=False),
            nn.BatchNorm2d(768),
            nn.ReLU(inplace=True))),
        ('flatten', nn.Flatten()),
        ('classifier', nn.Linear(768, num_classes))
    ]))


class FeatureExtractor(nn.Module):
    # CIFAR10:
    #   C: 36
    #   layer: 20
    #   dropout_p: 0.2
    # ImageNet:
    #   C: 48
    #   layer: 14
    #   dropout_p: None
    def __init__(self, genotype: Genotype, C: int = 36, layers: int = 20,
                 dropout_p: float = 0.2, std_conv: bool = False, **kwargs):
        super().__init__()
        self.genotype = genotype
        self.aux_C: int = 0
        self.aux_layer: int = 0

        stem_multiplier = 3
        C_curr = stem_multiplier * C
        self.stem = nn.Sequential(
            nn.Conv2d(3, C_curr, 3, padding=1, bias=False),
            nn.BatchNorm2d(C_curr)
        )

        C_prev_prev, C_prev, C_curr = C_curr, C_curr, C
        self.cells = nn.ModuleList()
        reduction_prev = False
        for i in range(layers):
            reduction = False
            if i in [layers // 3, 2 * layers // 3]:
                C_curr *= 2
                reduction = True
            cell = Cell(genotype, C_prev_prev, C_prev, C_curr,
                        reduction, reduction_prev, dropout_p=dropout_p,
                        std_conv=std_conv)
            reduction_prev = reduction
            self.cells.append(cell)
            C_prev_prev, C_prev = C_prev, cell.multiplier * C_curr
            if i == 2 * layers // 3:
                self.aux_C = C_prev
                self.aux_layer = i
                self.aux_dim = C_prev
        self.feats_dim = C_prev

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        s0 = s1 = self.stem(input)
        for cell in self.cells:
            s0, s1 = s1, cell(s0, s1)
        return s1

    def forward_with_aux(self, input: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        aux_feats: torch.Tensor = None
        s0 = s1 = self.stem(input)
        for i, cell in enumerate(self.cells):
            s0, s1 = s1, cell(s0, s1)
            if i == self.aux_layer:
                aux_feats = s1
        return s1, aux_feats



#!/usr/bin/env python3

from autovul.vision.configs import Config, config
from autovul.base.trainer import Trainer
import autovul.base.trainer

from typing import TYPE_CHECKING
from autovul.vision.datasets import ImageSet    # TODO: python 3.10
from autovul.vision.models import ImageModel
import argparse
if TYPE_CHECKING:
    pass


def add_argument(parser: argparse.ArgumentParser, ClassType: type[Trainer] = Trainer) -> argparse._ArgumentGroup:
    return autovul.base.trainer.add_argument(parser=parser, ClassType=ClassType)


def create(dataset_name: str = None, dataset: ImageSet = None, model: ImageModel = None,
           config: Config = config, ClassType=Trainer, **kwargs):
    return autovul.base.trainer.create(dataset_name=dataset_name, dataset=dataset,
                                    ClassType=ClassType,
                                    model=model, config=config, **kwargs)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import torch
import torch.nn as nn


class StdConv(nn.Module):
    def __init__(self, C_in, C_out):
        super(StdConv, self).__init__()
        self.conv = nn.Sequential(
            nn.Conv2d(C_in, C_out, 1, stride=1, padding=0, bias=False),
            nn.BatchNorm2d(C_out, affine=False),
            nn.ReLU()
        )

    def forward(self, x):
        return self.conv(x)


class PoolBranch(nn.Module):
    def __init__(self, pool_type, C_in, C_out, kernel_size, stride, padding, affine=False):
        super().__init__()
        self.preproc = StdConv(C_in, C_out)
        self.pool = Pool(pool_type, kernel_size, stride, padding)
        self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x):
        out = self.preproc(x)
        out = self.pool(out)
        out = self.bn(out)
        return out


class SeparableConv(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, stride, padding):
        super(SeparableConv, self).__init__()
        self.depthwise = nn.Conv2d(C_in, C_in, kernel_size=kernel_size, padding=padding, stride=stride,
                                   groups=C_in, bias=False)
        self.pointwise = nn.Conv2d(C_in, C_out, kernel_size=1, bias=False)

    def forward(self, x):
        out = self.depthwise(x)
        out = self.pointwise(out)
        return out


class ConvBranch(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, stride, padding, separable):
        super(ConvBranch, self).__init__()
        self.preproc = StdConv(C_in, C_out)
        if separable:
            self.conv = SeparableConv(C_out, C_out, kernel_size, stride, padding)
        else:
            self.conv = nn.Conv2d(C_out, C_out, kernel_size, stride=stride, padding=padding)
        self.postproc = nn.Sequential(
            nn.BatchNorm2d(C_out, affine=False),
            nn.ReLU()
        )

    def forward(self, x):
        out = self.preproc(x)
        out = self.conv(out)
        out = self.postproc(out)
        return out


class FactorizedReduce(nn.Module):
    def __init__(self, C_in, C_out, affine=False):
        super().__init__()
        self.conv1 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        self.conv2 = nn.Conv2d(C_in, C_out // 2, 1, stride=2, padding=0, bias=False)
        self.bn = nn.BatchNorm2d(C_out, affine=affine)

    def forward(self, x):
        out = torch.cat([self.conv1(x), self.conv2(x[:, :, 1:, 1:])], dim=1)
        out = self.bn(out)
        return out


class Pool(nn.Module):
    def __init__(self, pool_type, kernel_size, stride, padding):
        super().__init__()
        if pool_type.lower() == 'max':
            self.pool = nn.MaxPool2d(kernel_size, stride, padding)
        elif pool_type.lower() == 'avg':
            self.pool = nn.AvgPool2d(kernel_size, stride, padding, count_include_pad=False)
        else:
            raise ValueError()

    def forward(self, x):
        return self.pool(x)


class SepConvBN(nn.Module):
    def __init__(self, C_in, C_out, kernel_size, padding):
        super().__init__()
        self.relu = nn.ReLU()
        self.conv = SeparableConv(C_in, C_out, kernel_size, 1, padding)
        self.bn = nn.BatchNorm2d(C_out, affine=True)

    def forward(self, x):
        x = self.relu(x)
        x = self.conv(x)
        x = self.bn(x)
        return x



#!/usr/bin/env python3

import torch.cuda

from .tensor import *
from .others import *
from .environ import env
from .param import *


def empty_cache(threshold: float = None):
    threshold = threshold if threshold is not None else env['cache_threshold']
    if threshold is not None and env['num_gpus']:
        if torch.cuda.memory_cached() > threshold * (2**20):
            torch.cuda.empty_cache()



#!/usr/bin/env python3

from .imagemodel import ImageModel

from .nas import *
from .normal import *
from .others import *

from . import nas, normal, others

from autovul.vision.datasets import ImageSet
from autovul.vision.configs import Config, config
import autovul.base.models

import argparse
from typing import Union

module_list = [nas, normal, others]
__all__ = ['ImageModel', 'class_dict', 'add_argument', 'create',
           'get_available_models', 'output_available_models']
class_dict: dict[str, type[ImageModel]] = {}
for module in module_list:
    __all__.extend(module.__all__)
    class_dict.update(module.class_dict)


def add_argument(parser: argparse.ArgumentParser, model_name: str = None, model: Union[str, ImageModel] = None,
                 config: Config = config, class_dict: dict[str, type[ImageModel]] = class_dict):
    return autovul.base.models.add_argument(parser=parser, model_name=model_name, model=model,
                                         config=config, class_dict=class_dict)


def create(model_name: str = None, model: Union[str, ImageModel] = None,
           dataset_name: str = None, dataset: Union[str, ImageSet] = None,
           folder_path: str = None,
           config: Config = config, class_dict: dict[str, type[ImageModel]] = class_dict, **kwargs) -> ImageModel:
    return autovul.base.models.create(model_name=model_name, model=model,
                                   dataset_name=dataset_name, dataset=dataset,
                                   folder_path=folder_path,
                                   config=config, class_dict=class_dict, **kwargs)


def get_available_models(class_dict: dict[str, type[ImageModel]] = class_dict) -> dict[str, list[str]]:
    return autovul.base.models.get_available_models(class_dict=class_dict)


def output_available_models(class_dict: dict[str, type[ImageModel]] = class_dict, indent: int = 0) -> None:
    return autovul.base.models.output_available_models(class_dict=class_dict, indent=indent)



#!/usr/bin/env python3

# https://github.com/n-gao/pytorch-kfac/blob/master/torch_kfac/utils/lock.py

from contextlib import contextmanager


class Lock:
    r"""
    A boolean lock class used for contextmanager. It's used in :class:`~autovul.base.utils.fim.kfac.KFAC` to avoid auxiliary computation operations.

    * :attr:`__entered` specifies the current boolean status of the lock.

    Examples:
        >>> track = Lock()
        >>> print(bool(track))
        False
        >>> with track():
        >>>     print(bool(track))
        True
        >>> print(bool(track))
        False
    """

    def __init__(self) -> None:
        self.__entered: bool = False

    @contextmanager
    def __call__(self) -> None:
        assert not self.__entered
        try:
            self.__entered = True
            yield
        finally:
            self.__entered = False

    def __bool__(self) -> bool:
        return self.__entered

    def enable(self):
        """Set lock boolean value as `True`. It's used together with `Lock.disable()` when contextmanager is not suitable for the case."""
        self.__entered = True

    def disable(self):
        """Set lock boolean value as `False`. It's used together with `Lock.enable()` when contextmanager is not suitable for the case."""
        self.__entered = False



#!/usr/bin/env python3

from autovul.vision.datasets.imagefolder import ImageFolder

from torchvision.datasets import ImageNet as PytorchImageNet
import os

from autovul.vision import __file__ as root_file
root_dir = os.path.dirname(root_file)


class ImageNet(ImageFolder):

    name = 'imagenet'
    data_shape = [3, 224, 224]
    url = {
        'train': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_train.tar',
        'valid': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_val.tar',
        'test': 'http://www.image-net.org/challenges/LSVRC/2012/nnoupb/ILSVRC2012_img_test.tar',
    }
    md5 = {
        'train': '1d675b47d978889d74fa0da5fadfb00e',
        'valid': '29b22e2961454d5413ddabcf34fc5622',
    }

    def __init__(self, norm_par: dict[str, list[float]] = {'mean': [0.485, 0.456, 0.406],
                                                           'std': [0.229, 0.224, 0.225], },
                 **kwargs):
        super().__init__(norm_par=norm_par, **kwargs)

    def initialize_folder(self):
        try:
            PytorchImageNet(root=self.folder_path, split='train', download=True)
            PytorchImageNet(root=self.folder_path, split='val', download=True)
        except RuntimeError:
            raise RuntimeError('\n\n'
                               'You need to visit \'https://image-net.org/download-images.php\' '
                               'to download ImageNet.\n'
                               'There are direct links to files, but not legal to distribute. '
                               'Please apply for access permission and find links yourself.\n\n'
                               f'folder_path: {self.folder_path}\n'
                               'expected files:\n'
                               '{folder_path}/ILSVRC2012_devkit_t12.tar.gz\n'
                               '{folder_path}/ILSVRC2012_img_train.tar\n'
                               '{folder_path}/ILSVRC2012_img_val.tar\n'
                               '{folder_path}/meta.bin')
        os.symlink(os.path.join(self.folder_path, 'imagenet', 'val'),
                   os.path.join(self.folder_path, 'imagenet', 'valid'))



#!/usr/bin/env python3

from .trojannn import TrojanNN
from autovul.vision.utils.sgm import register_hook, remove_hook
from autovul.base.utils import AverageMeter, tanh_func

import torch
import torch.optim as optim
import argparse


class IMC(TrojanNN):

    r"""
    Input Model Co-optimization (IMC) Backdoor Attack is described in detail in the paper `A Tale of Evil Twins`_ by Ren Pang.

    Based on :class:`autovul.base.attacks.backdoor.BadNet`,
    IMC optimizes the watermark pixel values using PGD attack to enhance the performance.

    Args:
        target_value (float): The proportion of malicious images in the training set (Max 0.5). Default: 10.

    .. _A Tale of Evil Twins:
        https://arxiv.org/abs/1911.01559

    """

    name: str = 'imc'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--inner_iter', type=int)
        group.add_argument('--inner_lr', type=float)
        return group

    def __init__(self, inner_iter: int = 20, inner_lr: float = 0.1,
                 **kwargs):
        super().__init__(**kwargs)
        if self.mark.random_pos:
            raise Exception('IMC requires "random pos" to be False to train mark.')
        self.param_list['imc'] = ['inner_iter', 'inner_lr']
        self.inner_iter: int = inner_iter
        self.inner_lr: float = inner_lr

    def attack(self, epoch: int, **kwargs):
        super().attack(epoch, epoch_fn=self.epoch_fn, **kwargs)

    def epoch_fn(self, **kwargs):
        if self.model.sgm and 'sgm_remove' not in self.model.__dict__.keys():
            register_hook(self.model, self.model.sgm_gamma)
        self.optimize_mark()
        if self.model.sgm:
            remove_hook(self.model)

    def optimize_mark(self, loss_fn=None, **kwargs):
        atanh_mark = torch.randn_like(self.mark.mark) * self.mark.mask
        atanh_mark.requires_grad_()
        self.mark.mark = tanh_func(atanh_mark)
        optimizer = optim.Adam([atanh_mark], lr=self.inner_lr)
        optimizer.zero_grad()

        if loss_fn is None:
            loss_fn = self.model.loss

        losses = AverageMeter('Loss', ':.4e')
        for _epoch in range(self.inner_iter):
            for i, data in enumerate(self.dataset.loader['train']):
                if i > 20:
                    break
                _input, _label = self.model.get_data(data)
                poison_x = self.mark.add_mark(_input)
                loss = loss_fn(poison_x, self.target_class * torch.ones_like(_label))
                loss.backward()
                optimizer.step()
                optimizer.zero_grad()
                self.mark.mark = tanh_func(atanh_mark)
                losses.update(loss.item(), n=len(_label))
        atanh_mark.requires_grad = False
        self.mark.mark.detach_()



'''PNASNet in PyTorch.
Paper: Progressive Neural Architecture Search
'''

import torch
import torch.nn as nn
import torch.nn.functional as F

from typing import Union


class SepConv(nn.Module):
    '''Separable Convolution.'''

    def __init__(self, in_planes, out_planes, kernel_size, stride):
        super().__init__()
        self.conv1 = nn.Conv2d(in_planes, out_planes,
                               kernel_size, stride,
                               padding=(kernel_size - 1) // 2,
                               bias=False, groups=in_planes)
        self.bn1 = nn.BatchNorm2d(out_planes)

    def forward(self, x):
        return self.bn1(self.conv1(x))


class CellA(nn.Module):
    def __init__(self, in_planes, out_planes, stride=1):
        super().__init__()
        self.stride = stride
        self.sep_conv1 = SepConv(in_planes, out_planes, kernel_size=7, stride=stride)
        if stride == 2:
            self.conv1 = nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
            self.bn1 = nn.BatchNorm2d(out_planes)

    def forward(self, x):
        y1 = self.sep_conv1(x)
        y2 = F.max_pool2d(x, kernel_size=3, stride=self.stride, padding=1)
        if self.stride == 2:
            y2 = self.bn1(self.conv1(y2))
        return F.relu(y1 + y2)


class CellB(nn.Module):
    def __init__(self, in_planes, out_planes, stride=1):
        super().__init__()
        self.stride = stride
        # Left branch
        self.sep_conv1 = SepConv(in_planes, out_planes, kernel_size=7, stride=stride)
        self.sep_conv2 = SepConv(in_planes, out_planes, kernel_size=3, stride=stride)
        # Right branch
        self.sep_conv3 = SepConv(in_planes, out_planes, kernel_size=5, stride=stride)
        if stride == 2:
            self.conv1 = nn.Conv2d(in_planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
            self.bn1 = nn.BatchNorm2d(out_planes)
        # Reduce channels
        self.conv2 = nn.Conv2d(2 * out_planes, out_planes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn2 = nn.BatchNorm2d(out_planes)

    def forward(self, x):
        # Left branch
        y1 = self.sep_conv1(x)
        y2 = self.sep_conv2(x)
        # Right branch
        y3 = F.max_pool2d(x, kernel_size=3, stride=self.stride, padding=1)
        if self.stride == 2:
            y3 = self.bn1(self.conv1(y3))
        y4 = self.sep_conv3(x)
        # Concat & reduce channels
        b1 = F.relu(y1 + y2)
        b2 = F.relu(y3 + y4)
        y = torch.cat([b1, b2], 1)
        return F.relu(self.bn2(self.conv2(y)))


class PNASNet(nn.Module):
    def __init__(self, cell_type: Union[type[CellA], type[CellB]],
                 num_planes: int, num_cells: int = 6, num_classes: int = 10):
        super().__init__()
        self.in_planes = num_planes
        self.cell_type = cell_type

        self.conv1 = nn.Conv2d(3, num_planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(num_planes)

        self.layer1 = self._make_layer(num_planes, num_cells=num_cells)
        self.layer2 = self._downsample(num_planes * 2)
        self.layer3 = self._make_layer(num_planes * 2, num_cells=num_cells)
        self.layer4 = self._downsample(num_planes * 4)
        self.layer5 = self._make_layer(num_planes * 4, num_cells=num_cells)

        self.linear = nn.Linear(num_planes * 4, num_classes)

    def _make_layer(self, planes, num_cells):
        layers = []
        for _ in range(num_cells):
            layers.append(self.cell_type(self.in_planes, planes, stride=1))
            self.in_planes = planes
        return nn.Sequential(*layers)

    def _downsample(self, planes):
        layer = self.cell_type(self.in_planes, planes, stride=2)
        self.in_planes = planes
        return layer

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = self.layer5(out)
        out = F.avg_pool2d(out, 8)
        out = self.linear(out.view(out.size(0), -1))
        return out


def PNASNetA(**kwargs):
    return PNASNet(CellA, num_cells=6, num_planes=44, **kwargs)


def PNASNetB(**kwargs):
    return PNASNet(CellB, num_cells=6, num_planes=32, **kwargs)



#!/usr/bin/env python3

from autovul.vision.datasets.imageset import ImageSet
from autovul.vision.models.imagemodel import ImageModel
from autovul.base.attacks import Attack
from autovul.base.utils.data import TensorListDataset, dataset_to_list


import torch
import math
import random
import os
import argparse
from typing import Callable
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import torch.utils.data


class PoisonRandom(Attack):
    name: str = 'poison_random'

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--poison_percent', type=float,
                           help='malicious training data injection probability for each batch, defaults to 0.01')
        group.add_argument('--train_mode', choices=['batch', 'dataset'],
                           help='target class of backdoor, defaults to "batch"')
        return group

    def __init__(self, poison_percent: float = 0.01, train_mode: str = 'batch', **kwargs):
        super().__init__(**kwargs)
        self.dataset: ImageSet
        self.model: ImageModel
        self.param_list['badnet'] = ['poison_percent']
        self.poison_percent: float = poison_percent
        self.poison_num = self.dataset.batch_size * self.poison_percent
        self.train_mode: str = train_mode

    def attack(self, epoch: int, save=False, **kwargs):
        if self.train_mode == 'batch':
            self.model._train(epoch, save=save,
                              validate_fn=self.validate_fn, get_data_fn=self.get_data,
                              save_fn=self.save, **kwargs)
        elif self.train_mode == 'dataset':
            dataset = self.mix_dataset()
            loader = self.dataset.get_dataloader('train', dataset=dataset)
            self.model._train(epoch, save=save,
                              validate_fn=self.validate_fn, loader_train=loader,
                              save_fn=self.save, **kwargs)

    def mix_dataset(self) -> torch.utils.data.Dataset:
        clean_dataset = self.dataset.loader['train'].dataset
        subset, otherset = ImageSet.split_dataset(clean_dataset, percent=self.poison_percent)
        if not len(subset):
            return clean_dataset
        _input, _label = dataset_to_list(subset)
        _input = torch.stack(_input)

        _label = torch.tensor(_label, dtype=torch.long)
        _label += torch.randint_like(_label, low=1, high=self.model.num_classes)
        _label %= self.model.num_classes
        _label = _label.tolist()
        poison_dataset = TensorListDataset(_input, _label)
        return torch.utils.data.ConcatDataset([poison_dataset, otherset])

    # ---------------------- I/O ----------------------------- #

    def get_filename(self, **kwargs):
        return f'{self.train_mode}_{self.poison_percent}'

    def save(self, filename: str = None, **kwargs):
        filename = filename if filename is not None else self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.model.save(file_path + '.pth')
        print('attack results saved at: ', file_path)

    def load(self, filename: str = None, **kwargs):
        filename = filename if filename is not None else self.get_filename(**kwargs)
        file_path = os.path.join(self.folder_path, filename)
        self.model.load(file_path + '.pth')
        print('attack results loaded from: ', file_path)

    # ---------------------- Utils ---------------------------- #

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        _input, _label = self.model.get_data(data)
        decimal, integer = math.modf(self.poison_num)
        integer = int(integer)
        if random.uniform(0, 1) < decimal:
            integer += 1
        if integer:
            _label[:integer] += torch.randint_like(_label[:integer], low=1, high=self.model.num_classes)
            _label[:integer] %= self.model.num_classes
        return _input, _label

    def validate_fn(self, get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
                    indent: int = 0, **kwargs) -> tuple[float, float]:
        return self.model._validate(indent=indent, **kwargs)



#!/usr/bin/env python3

from autovul.base.models import Model
from autovul.base.configs import config
from autovul.base.utils import get_name
from autovul.base.utils.output import ansi, prints
from autovul.base.utils.fim import KFAC, EKFAC

from typing import TYPE_CHECKING
from typing import Any, Union
from autovul.base.configs import Config    # TODO: python 3.10
from autovul.base.datasets import Dataset
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
if TYPE_CHECKING:
    pass


class Trainer:
    name = 'trainer'
    param_list = ['optim_args', 'train_args', 'writer_args',
                  'optimizer', 'lr_scheduler', 'pre_conditioner', 'writer']

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        group.add_argument('--epoch', type=int, help='training epochs, defaults to config[train][epoch].')
        group.add_argument('--resume', type=int, help='resume training from certain epoch, defaults to be 0.')
        group.add_argument('--lr', type=float, help='learning rate, defaults to 0.1.')
        group.add_argument('--parameters', default='full',
                           help='training parameters ("features", "classifier", "full"), defaults to "full".')
        group.add_argument('--OptimType', help='optimizer type, defaults to SGD.')
        group.add_argument('--momentum', type=float, help='momentum passed to Optimizer, defaults to 0.9.')
        group.add_argument('--weight_decay', type=float, help='weight_decay passed to Optimizer, defaults to 3e-4.')
        group.add_argument('--nesterov', action='store_true', help='enable nesterov for SGD optimizer.')
        group.add_argument('--lr_scheduler', action='store_true', help='enable CosineAnnealingLR scheduler.')
        group.add_argument('--pre_conditioner', choices=['kfac', 'ekfac'], help='Using KFAC preconditioner.')
        group.add_argument('--amp', action='store_true', help='Automatic Mixed Precision.')
        group.add_argument('--grad_clip', type=float, help='Gradient Clipping max norms.')
        group.add_argument('--validate_interval', type=int,
                           help='validate interval during training epochs, defaults to 10.')
        group.add_argument('--save', action='store_true', help='save training results.')
        group.add_argument('--tensorboard', action='store_true', help='save training logging for tensorboard.')
        group.add_argument('--log_dir', help='save training logging for tensorboard.')
        group.add_argument('--flush_secs', type=int,
                           help='How often, in seconds, to flush the pending events and summaries to disk.')
        return group

    def __init__(self, optim_args: dict[str, Any] = {}, train_args: dict[str, Any] = {},
                 writer_args: dict[str, Any] = {},
                 optimizer: Optimizer = None, lr_scheduler: _LRScheduler = None,
                 pre_conditioner: Union[KFAC, EKFAC] = None, writer=None):
        self.optim_args = optim_args.copy()   # TODO: issue 6 why? to avoid BadAppend issues
        self.train_args = train_args.copy()
        self.writer_args = writer_args.copy()
        self.optimizer = optimizer
        self.lr_scheduler = lr_scheduler
        self.pre_conditioner = pre_conditioner
        self.writer = writer

    def __getitem__(self, key: str):
        if key in self.train_args.keys():
            return self.train_args[key]
        return getattr(self, key)

    def __getattr__(self, key: str):
        if key in self.train_args.keys():
            return self.train_args[key]
        raise AttributeError(key)

    def keys(self):
        keys = self.param_list.copy()
        keys.remove('optim_args')
        keys.remove('train_args')
        keys.remove('writer_args')
        keys.extend(self.train_args.keys())
        return keys

    def summary(self, indent: int = 0):
        prints('{blue_light}{0:<30s}{reset} Parameters: '.format(self.name, **ansi), indent=indent)
        prints(self.__class__.__name__, indent=indent)
        for key in self.param_list:
            value = getattr(self, key)
            if value:
                prints('{green}{0:<10s}{reset}'.format(key, **ansi), indent=indent + 10)
                prints(value, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)


def add_argument(parser: argparse.ArgumentParser, ClassType: type[Trainer] = Trainer):
    group = parser.add_argument_group('{yellow}trainer{reset}'.format(**ansi))
    return ClassType.add_argument(group)


def create(dataset_name: str = None, dataset: Dataset = None, model: Model = None,
           pre_conditioner: str = None, ClassType: type[Trainer] = Trainer, tensorboard: bool = None,
           config: Config = config, **kwargs):
    assert isinstance(model, Model)
    dataset_name = get_name(name=dataset_name, module=dataset, arg_list=['-d', '--dataset'])
    result = config.get_config(dataset_name=dataset_name)['trainer'].update(kwargs)

    optim_keys = model.define_optimizer.__code__.co_varnames
    train_keys = model._train.__code__.co_varnames
    optim_args: dict[str, Any] = {}
    train_args: dict[str, Any] = {}
    for key, value in result.items():
        if key in optim_keys:
            _dict = optim_args
        elif key in train_keys and key != 'verbose':
            _dict = train_args
        else:
            continue
        _dict[key] = value
    optimizer, lr_scheduler = model.define_optimizer(T_max=result['epoch'], **optim_args)

    module = model._model
    if optim_args['parameters'] == 'features':
        module = module.features
    elif optim_args['parameters'] == 'classifier':
        module = module.classifier
    kfac_optimizer = None
    if pre_conditioner == 'kfac':
        kfac_optimizer = KFAC(module)
    elif pre_conditioner == 'ekfac':
        kfac_optimizer = EKFAC(module)

    writer = None
    writer_args: dict[str, Any] = {}
    if tensorboard:
        from torch.utils.tensorboard import SummaryWriter
        writer_keys = SummaryWriter.__init__.__code__.co_varnames   # log_dir, flush_secs, ...
        for key, value in result.items():
            if key in writer_keys:
                writer_args[key] = value
        writer = SummaryWriter(**writer_args)
    return ClassType(optim_args=optim_args, train_args=train_args, writer_args=writer_args,
                     optimizer=optimizer, lr_scheduler=lr_scheduler, pre_conditioner=kfac_optimizer, writer=writer)



#!/usr/bin/env python3

import autovul.base.configs
from autovul.base.configs import Config

import os

config_path: dict[str, str] = {
    'package': os.path.dirname(__file__),
    'user': os.path.normpath(os.path.expanduser('~/.autovul/configs/vision')),
    'project': os.path.normpath('./configs/vision'),
}
config = Config(_base=autovul.base.configs.config, **config_path)



#!/usr/bin/env python3

__all__ = ['to_tensor', 'to_numpy', 'to_list']

from autovul.vision import environ as environ
from autovul.vision import datasets as datasets
from autovul.vision import models as models
from autovul.vision import trainer as trainer
from autovul.vision import attacks as attacks
from autovul.vision import marks as marks

from autovul.base.utils import to_tensor, to_numpy, to_list



#!/usr/bin/env python3

from .others import BasicObject
from .output import prints, output_iter

import os
from collections.abc import Iterable

from typing import TYPE_CHECKING
from typing import Union    # TODO: python 3.10
from autovul.base.datasets import Dataset    # TODO: python 3.10
from autovul.base.models import Model
if TYPE_CHECKING:
    pass


class Process(BasicObject):
    name: str = 'process'

    def __init__(self, output: Union[int, Iterable[str]] = 0, indent: int = 0, **kwargs):
        super().__init__(**kwargs)
        self.param_list['verbose'] = ['output', 'indent']

        self.output: set[str] = None
        self.output = self.get_output(output)
        self.indent = indent

    # -----------------------------------Output-------------------------------------#
    def summary(self, indent: int = None):
        indent = indent if indent is not None else self.indent
        return super().summary(indent=indent)

    def get_output(self, org_output: Union[int, Iterable[str]] = None) -> set[str]:
        if org_output is None:
            return self.output
        elif isinstance(org_output, int):
            return self.get_output_int(org_output)
        return set(org_output)

    @classmethod
    def get_output_int(cls, org_output: int = 0) -> set[str]:
        result: set[str] = set()
        if org_output >= 5:
            result.add('end')
        if org_output >= 10:
            result.add('start')
        if org_output >= 20:
            result.add('middle')
        if org_output >= 30:
            result.add('memory')
        return result

    @staticmethod
    def output_iter(name: str, _iter: int, iteration: int = None, indent: int = 0):
        prints(f'{name} Iter: {output_iter(_iter + 1, iteration)}', indent=indent)


class Model_Process(Process):
    name: str = 'model_process'

    def __init__(self, dataset: Dataset = None, model: Model = None, folder_path: str = None, **kwargs):
        super().__init__(**kwargs)
        self.param_list['process'] = ['clean_acc', 'folder_path']
        self.dataset = dataset
        self.model = model

        if folder_path is not None:
            folder_path = os.path.normpath(folder_path)
            if not os.path.exists(folder_path):
                os.makedirs(folder_path)
        self.folder_path = folder_path
        self.__clean_acc: float = None

    @property
    def clean_acc(self) -> float:
        if self.__clean_acc is None:
            _, self.__clean_acc = self.model._validate(verbose=False)
        return self.__clean_acc



#!/usr/bin/env python3

from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torch
import torch.nn as nn
from collections import OrderedDict
import os


class FeatureExtractor(nn.Module):
    def __init__(self, model: nn.Module, **kwargs):
        super().__init__()
        self.num_layers: int = model.num_layers
        self.pool_layers_idx: int = model.pool_layers_idx
        self.stem: nn.Sequential = model.stem
        self.layers: nn.ModuleList = model.layers
        self.pool_layers: nn.ModuleList = model.pool_layers

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        cur = self.stem(x)
        layers = [cur]
        for layer_id in range(self.num_layers):
            cur = self.layers[layer_id](layers)
            layers.append(cur)
            if layer_id in self.pool_layers_idx:
                for i, layer in enumerate(layers):
                    layers[i] = self.pool_layers[self.pool_layers_idx.index(layer_id)](layer)
                cur = layers[-1]
        return cur


class _ENAS(_ImageModel):
    def __init__(self, _model: nn.Module, **kwargs):
        super().__init__(**kwargs)
        self.features = FeatureExtractor(_model)
        self.classifier = nn.Sequential(OrderedDict([
            ('dropout', _model.dropout),
            ('dense', _model.dense),
        ]))


class ENAS(ImageModel):
    available_models = ['enas']

    def __init__(self, name: str = 'enas', model: type[_ENAS] = _ENAS, folder_path: str = None, **kwargs):
        import sys
        from autovul.vision.utils.model_archs.enas import __file__ as file_path
        sys.path.append(os.path.dirname(file_path))
        _model = torch.load(os.path.join(folder_path, 'enas_macro.pt'))  # generated by nni library
        super().__init__(name=name, model=model, _model=_model, folder_path=folder_path, **kwargs)



#!/usr/bin/env python3

__all__ = ['BadNet', 'TrojanNN', 'LatentBackdoor', 'IMC']

from .badnet import BadNet
from .trojannn import TrojanNN
from .latent_backdoor import LatentBackdoor
from .imc import IMC



#!/usr/bin/env python3

import autovul.vision
from autovul.vision.utils import summary, to_numpy

import torch
import numpy as np
import argparse

import time

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    autovul.vision.environ.add_argument(parser)
    autovul.vision.datasets.add_argument(parser)
    autovul.vision.models.add_argument(parser)
    args = parser.parse_args()

    env = autovul.vision.environ.create(**args.__dict__)
    dataset = autovul.vision.datasets.create(**args.__dict__)
    model = autovul.vision.models.create(dataset=dataset, **args.__dict__)

    if env['verbose']:
        summary(env=env, dataset=dataset, model=model)
    # loss, acc1 = model._validate()

    model.activate_params(model.parameters())
    model.zero_grad()

    torch.random.manual_seed(int(time.time()))
    grad_x = None
    grad_xx = None
    n_sample = 512

    for i, data in enumerate(dataset.get_dataloader('valid', shuffle=True, batch_size=1, drop_last=True)):
        if i >= n_sample:
            break
        _input, _label = model.get_data(data)
        loss = model.loss(_input, _label)
        loss.backward()
        grad_temp_list = []
        for param in model.parameters():
            grad_temp_list.append(param.grad.flatten())
        grad = torch.cat(grad_temp_list)
        grad = grad if grad.norm(p=2) <= 5.0 else grad / grad.norm(p=2) * 5.0
        grad_temp = grad.detach().cpu().clone()
        if grad_x is None:
            grad_x = grad_temp / n_sample
            grad_xx = grad_temp.square() / n_sample
        else:
            grad_x += grad_temp / n_sample
            grad_xx += grad_temp.square() / n_sample
        model.zero_grad()

    model.eval()
    model.activate_params([])

    grad_tensor = to_numpy(grad_xx - grad_x.square())
    grad_tensor[grad_tensor < 0] = 0
    var = float(np.sum(np.sqrt(grad_tensor)))

    print(f'{model.name:20}  {var:f}')



#!/usr/bin/env python3

from autovul.vision.configs import Config, config
from autovul.vision.datasets import ImageSet
from autovul.vision.environ import env
from autovul.base.utils import to_tensor, to_numpy, byte2float, gray_img, save_tensor_as_img
from autovul.base.utils.output import ansi, prints, Indent_Redirect

import os
import sys
import random
import numpy as np
import torch
import argparse
import PIL.Image as Image
from collections.abc import Callable
from typing import Union

redirect = Indent_Redirect(buffer=True, indent=0)
dir_path = os.path.dirname(__file__)


def add_argument(parser: argparse.ArgumentParser):
    group = parser.add_argument_group('{yellow}mark{reset}'.format(**ansi))
    return Watermark.add_argument(group)


def create(mark_path: str = None, data_shape: list[int] = None,
           dataset_name: str = None, dataset: Union[str, ImageSet] = None,
           config: Config = config, **kwargs):
    if data_shape is None:
        assert isinstance(dataset, ImageSet)
        data_shape = dataset.data_shape
    if dataset_name is None and dataset is not None:
        dataset_name = dataset.name
    result = config.get_config(dataset_name=dataset_name)['mark'].update(kwargs).update(mark_path=mark_path)
    return Watermark(data_shape=data_shape, **result)


class Watermark:
    name: str = 'mark'

    @staticmethod
    def add_argument(group: argparse._ArgumentGroup):
        group.add_argument('--edge_color', help='edge color in watermark image, defaults to "auto".')
        group.add_argument('--mark_path', help='edge color in watermark image, '
                           'defaults to "autovul.base/data/mark/square_white.png".')
        group.add_argument('--mark_alpha', type=float, help='mark transparency, defaults to 0.0.')
        group.add_argument('--mark_height', type=int, help='mark height, defaults to 3.')
        group.add_argument('--mark_width', type=int, help='mark width, defaults to 3.')
        group.add_argument('--height_offset', type=int, help='height offset, defaults to 0')
        group.add_argument('--width_offset', type=int, help='width offset, defaults to 0')
        group.add_argument('--random_pos', action='store_true', help='Random offset Location for add_mark.')
        group.add_argument('--random_init', action='store_true', help='random values for mark pixel.')
        group.add_argument('--mark_distributed', action='store_true', help='Distributed Mark.')
        return group

    def __init__(self, mark_path: str = 'autovul.base/data/mark/square_white.png',
                 data_shape: list[int] = None, edge_color: Union[str, torch.Tensor] = 'auto',
                 mark_alpha: float = 0.0, mark_height: int = None, mark_width: int = None,
                 height_offset: int = 0, width_offset: int = 0,
                 random_pos=False, random_init=False, mark_distributed=False,
                 add_mark_fn=None, **kwargs):
        self.param_list: dict[str, list[str]] = {}
        self.param_list['mark'] = ['mark_path', 'data_shape', 'edge_color',
                                   'mark_alpha', 'mark_height', 'mark_width',
                                   'random_pos', 'random_init']
        assert mark_height > 0 and mark_width > 0
        # --------------------------------------------------- #

        # WaterMark Image Parameters
        self.mark_alpha: float = mark_alpha
        self.data_shape: list[int] = data_shape
        self.mark_path: str = mark_path
        self.mark_height: int = mark_height
        self.mark_width: int = mark_width
        self.random_pos = random_pos
        self.random_init = random_init
        self.mark_distributed = mark_distributed
        self.add_mark_fn: Callable = add_mark_fn
        # --------------------------------------------------- #

        if self.mark_distributed:
            self.mark = torch.rand(data_shape, dtype=torch.float, device=env['device'])
            mask = torch.zeros(data_shape[-2:], dtype=torch.bool, device=env['device']).flatten()
            idx = np.random.choice(len(mask), self.mark_height * self.mark_width, replace=False).tolist()
            mask[idx] = 1.0
            mask = mask.view(data_shape[-2:])
            self.mask = mask
            self.alpha_mask = self.mask * (1 - mark_alpha)
            self.edge_color = None
        else:
            org_mark_img: Image.Image = self.load_img(img_path=mark_path,
                                                      height=mark_height, width=mark_width, channel=data_shape[0])
            self.org_mark: torch.Tensor = byte2float(org_mark_img)
            self.edge_color: torch.Tensor = self.get_edge_color(
                self.org_mark, data_shape, edge_color)
            self.org_mask, self.org_alpha_mask = self.org_mask_mark(self.org_mark, self.edge_color, self.mark_alpha)
            if random_init:
                self.org_mark = self.random_init_mark(self.org_mark, self.org_mask)
            if not random_pos:
                self.param_list['mark'].extend(['height_offset', 'width_offset'])
                self.height_offset: int = height_offset
                self.width_offset: int = width_offset
                self.mark, self.mask, self.alpha_mask = self.mask_mark()

    # add mark to the Image with mask.

    def add_mark(self, _input: torch.Tensor, random_pos=None, alpha: float = None, **kwargs) -> torch.Tensor:
        if callable(self.add_mark_fn):
            return self.add_mark_fn(_input, random_pos=random_pos, alpha=alpha, **kwargs)
        if random_pos is None:
            random_pos = self.random_pos
        if random_pos:
            # batch_size = _input.size(0)
            # height_offset = torch.randint(high=self.data_shape[-2] - self.mark_height, size=[batch_size])
            # width_offset = torch.randint(high=self.data_shape[-1] - self.mark_width, size=[batch_size])
            height_offset = random.randint(0, self.data_shape[-2] - self.mark_height)
            width_offset = random.randint(0, self.data_shape[-1] - self.mark_width)
            mark, mask, alpha_mask = self.mask_mark(height_offset=height_offset, width_offset=width_offset)
        else:
            mark, mask, alpha_mask = self.mark, self.mask, self.alpha_mask
            if alpha is not None:
                alpha_mask = torch.ones_like(self.alpha_mask) * (1 - alpha)
        _mask = mask * alpha_mask
        mark, _mask = mark.to(_input.device), _mask.to(_input.device)
        return _input + _mask * (mark - _input)

    @staticmethod
    def get_edge_color(mark: torch.Tensor, data_shape: list[int],
                       edge_color: Union[str, torch.Tensor] = 'auto') -> torch.Tensor:

        assert data_shape[0] == mark.shape[0]
        t: torch.Tensor = torch.zeros(data_shape[0], dtype=torch.float)
        if isinstance(edge_color, str):
            if edge_color == 'black':
                pass
            elif edge_color == 'white':
                t += 1
            elif edge_color == 'auto':
                mark = mark.transpose(0, -1)
                if mark.flatten(start_dim=1).std(dim=1).max() < 1e-3:
                    t = -torch.ones_like(mark[0, 0])
                else:
                    _list = [mark[0, :, :], mark[-1, :, :],
                             mark[:, 0, :], mark[:, -1, :]]
                    _list = torch.cat(_list)
                    t = _list.mode(dim=0)[0]
            else:
                raise ValueError(edge_color)
        else:
            t = torch.as_tensor(edge_color)
            assert t.dim() == 1
            assert t.shape[0] == data_shape[0]
        return t

    @staticmethod
    def org_mask_mark(org_mark: torch.Tensor, edge_color: torch.Tensor, mark_alpha: float) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        height, width = org_mark.shape[-2:]
        mark = torch.zeros_like(org_mark, dtype=torch.float)
        mask = torch.zeros([height, width], dtype=torch.bool)
        for i in range(height):
            for j in range(width):
                if not org_mark[:, i, j].equal(edge_color):
                    mark[:, i, j] = org_mark[:, i, j]
                    mask[i, j] = 1
        alpha_mask = mask * (1 - mark_alpha)
        return mask, alpha_mask

    def mask_mark(self, org_mark: torch.Tensor = None, org_mask: torch.Tensor = None, org_alpha_mask: torch.Tensor = None,
                  height_offset: int = None, width_offset: int = None) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        if org_mark is None:
            org_mark = self.org_mark
        if org_mask is None:
            org_mask = self.org_mask
        if org_alpha_mask is None:
            org_alpha_mask = self.org_alpha_mask
        if height_offset is None:
            height_offset = self.height_offset
        if width_offset is None:
            width_offset = self.width_offset
        mark = -torch.ones(self.data_shape, dtype=torch.float)
        mask = torch.zeros(self.data_shape[-2:], dtype=torch.bool)
        alpha_mask = torch.zeros_like(mask, dtype=torch.float)

        start_h = height_offset
        start_w = width_offset
        end_h = height_offset + self.mark_height
        end_w = width_offset + self.mark_width

        mark[:, start_h:end_h, start_w:end_w] = org_mark
        mask[start_h:end_h, start_w:end_w] = org_mask
        alpha_mask[start_h:end_h, start_w:end_w] = org_alpha_mask
        if env['num_gpus']:
            mark = mark.to(env['device'])
            mask = mask.to(env['device'])
            alpha_mask = alpha_mask.to(env['device'])
        return mark, mask, alpha_mask

    """
    # each image in the batch has a unique random location.
    def mask_mark_batch(self, height_offset: torch.Tensor, width_offset: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        assert len(height_offset) == len(width_offset)
        shape = [len(height_offset)].extend(self.data_shape)
        mark = -torch.ones(shape, dtype=int)
        shape[1] = 1
        mask = torch.zeros(shape, dtype=torch.float)
        alpha_mask = torch.zeros_like(mask)

        start_h = height_offset
        start_w = width_offset
        end_h = height_offset + self.mark_height
        end_w = width_offset + self.mark_width

        mark[:, start_h:end_h, start_w:end_w] = self.org_mark
        mask[start_h:end_h, start_w:end_w] = self.org_mask
        alpha_mask[start_h:end_h, start_w:end_w] = self.org_alpha_mask

        mark = to_tensor(mark)
        mask = to_tensor(mask)
        alpha_mask = to_tensor(alpha_mask)
        return mark, mask, alpha_mask
    """

    # Give the mark init values for non transparent pixels.
    @staticmethod
    def random_init_mark(mark, mask):
        init_mark = torch.rand_like(mark)
        ones = -torch.ones_like(mark)
        init_mark = torch.where(mask, init_mark, ones)
        return init_mark

    # ------------------------------ I/O --------------------------- #

    @staticmethod
    def load_img(img_path: str, height: int, width: int, channel: int = 3) -> Image.Image:
        if not os.path.exists(img_path) and not os.path.exists(img_path := os.path.join(dir_path, img_path)):
            raise FileNotFoundError(img_path.removeprefix(dir_path))
        mark: Image.Image = Image.open(img_path)
        mark = mark.resize((width, height), Image.ANTIALIAS)

        if channel == 1:
            mark = gray_img(mark, num_output_channels=1)
        elif channel == 3 and mark.mode in ['1', 'L']:
            mark = gray_img(mark, num_output_channels=3)
        return mark

    def save_img(self, img_path: str):
        img = self.org_mark * self.org_mask if self.random_pos else self.mark * self.mask
        save_tensor_as_img(img_path, img)

    def load_npz(self, npz_path: str):
        if not os.path.exists(npz_path) and not os.path.exists(npz_path := os.path.join(dir_path, npz_path)):
            raise FileNotFoundError(npz_path.removeprefix(dir_path))
        _dict = np.load(npz_path)
        if not self.mark_distributed:
            self.org_mark = torch.as_tensor(_dict['org_mark'])
            self.org_mask = torch.as_tensor(_dict['org_mask'])
            self.org_alpha_mask = torch.as_tensor(_dict['org_alpha_mask'])
        if not self.random_pos:
            self.mark = to_tensor(_dict['mark'])
            self.mask = to_tensor(_dict['mask'])
            self.alpha_mask = to_tensor(_dict['alpha_mask'])

    def save_npz(self, npz_path: str):
        _dict = {}
        if not self.mark_distributed:
            _dict |= {'org_mark': to_numpy(self.org_mark),
                      'org_mask': to_numpy(self.org_mask),
                      'org_alpha_mask': to_numpy(self.org_alpha_mask)}
        if not self.random_pos:
            _dict |= {'mark': to_numpy(self.mark),
                      'mask': to_numpy(self.mask),
                      'alpha_mask': to_numpy(self.alpha_mask)}
        np.savez(npz_path, **_dict)

    # ------------------------------Verbose Information--------------------------- #
    def summary(self, indent: int = 0):
        prints('{blue_light}{0:<30s}{reset} Parameters: '.format(self.name, **ansi), indent=indent)
        prints(self.__class__.__name__, indent=indent)
        for key, value in self.param_list.items():
            if value:
                prints('{green}{0:<20s}{reset}'.format(key, **ansi), indent=indent + 10)
                prints({v: getattr(self, v) for v in value}, indent=indent + 10)
                prints('-' * 20, indent=indent + 10)

    def __str__(self) -> str:
        sys.stdout = redirect
        self.summary()
        _str = redirect.buffer
        redirect.reset()
        return _str



#!/usr/bin/env python3

from autovul.base.utils import *

import torch
from typing import Union

from matplotlib.colors import Colormap  # type: ignore  # TODO
from matplotlib.cm import get_cmap  # type: ignore  # TODO
jet = get_cmap('jet')
# jet = torch.tensor(jet(np.arange(256))[:, :3])


def apply_cmap(heatmap: torch.Tensor, cmap: Union[Colormap, torch.Tensor] = jet) -> torch.Tensor:
    if cmap is None:
        return heatmap
    heatmap = heatmap.detach().cpu()
    squeeze_flag = False
    if heatmap.dim() == 2:
        heatmap = heatmap.unsqueeze(0)  # (N, H, W)
        squeeze_flag = True
    if isinstance(cmap, Colormap):      # Note that C==4 for most cmaps
        heatmap = torch.as_tensor(cmap(heatmap.numpy()))  # (N, H, W, C)
    else:
        cmap = torch.as_tensor(cmap)
        assert cmap.shape[0] == 256     # cmap: [256, 3|4]
        heatmap = cmap[(heatmap * 255).long()]  # (N, H, W, C)  uint8
    heatmap = heatmap.permute(0, 3, 1, 2).contiguous().float()  # (N, C, H, W)
    if heatmap.max() > 1:
        heatmap.div_(255)
    return heatmap[0] if squeeze_flag else heatmap


def superimpose(foreground: torch.Tensor, background: torch.Tensor, alpha: float = 1.0):
    foreground = foreground.to(device=background.device)
    assert foreground.shape[-2:] == background.shape[-2:]
    squeeze_flag = False
    if foreground.dim() == 3 and background.dim() == 3:
        squeeze_flag = True
    alpha_flag = True
    if foreground.shape == 3:
        foreground = foreground.unsqueeze(0)  # (N, C, H, W)
    if background.shape == 3:
        background = background.unsqueeze(0)  # (N, C, H, W)
    if background.shape[1] == 3:
        alpha_flag = False
        shape = list(background.shape)
        shape[1] = 1
        alpha_channel = torch.ones(shape, device=background.device)
        background = torch.cat((background, alpha_channel), dim=1)  # (N, 4, H, W)
    if foreground.shape[1] == 3:
        shape = list(foreground.shape)
        shape[1] = 1
        alpha_channel = torch.ones(shape, device=foreground.device)
        foreground = torch.cat((foreground, alpha_channel), dim=1)  # (N, 4, H, W)
    foreground[:, 3] *= alpha
    mix = _superimpose(foreground, background)
    mix = mix[:, :3] if alpha_flag else mix
    return mix[0] if squeeze_flag else mix


def _superimpose(foreground: torch.Tensor, background: torch.Tensor):
    # (N, 4, H, W)
    mix = torch.empty_like(background)
    mix[:, 3] = 1 - (1 - foreground[:, 3]) * (1 - background[:, 3])  # alpha
    mix[:, :3] = foreground[:, :3] * foreground[:, 3, None] + \
        background[:, :3] * background[:, 3, None] * (1 - foreground[:, 3, None])
    mix[:, :3] /= mix[:, 3, None]
    return mix



#!/usr/bin/env python3

import torch
import torch.nn as nn
import torch.nn.functional as F


def _compute_zero_padding(kernel_size: int) -> int:
    """Computes zero padding."""
    return (kernel_size - 1) // 2


class SSIM(nn.Module):
    r"""Creates a criterion that measures the Structural Similarity (SSIM)
    index between each element in the input `x` and target `y`.

    The index can be described as:

    .. math::

      \text{SSIM}(x, y) = \frac{(2\mu_x\mu_y+c_1)(2\sigma_{xy}+c_2)}
      {(\mu_x^2+\mu_y^2+c_1)(\sigma_x^2+\sigma_y^2+c_2)}

    where:
      - :math:`c_1=(k_1 L)^2` and :math:`c_2=(k_2 L)^2` are two variables to
        stabilize the division with weak denominator.
      - :math:`L` is the dynamic range of the pixel-values (typically this is
        :math:`2^{\#\text{bits per pixel}}-1`).

    the loss, or the Structural dissimilarity (DSSIM) can be finally described
    as:

    .. math::

      \text{loss}(x, y) = \frac{1 - \text{SSIM}(x, y)}{2}

    Arguments:
        window_size (int): the size of the kernel.
        max_val (float): the dynamic range of the images. Default: 1.
        reduction (str, optional): Specifies the reduction to apply to the
         output: 'none' | 'mean' | 'sum'. 'none': no reduction will be applied,
         'mean': the sum of the output will be divided by the number of elements
         in the output, 'sum': the output will be summed. Default: 'none'.

    Returns:
        Tensor: the ssim index.

    Shape:
        - Input: :math:`(B, C, H, W)`
        - Target :math:`(B, C, H, W)`
        - Output: scale, if reduction is 'none', then :math:`(B, C, H, W)`

    # Examples::

    #     >>> input1 = torch.rand(1, 4, 5, 5)
    #     >>> input2 = torch.rand(1, 4, 5, 5)
    #     >>> ssim = kornia.losses.SSIM(5, reduction='none')
    #     >>> loss = ssim(input1, input2)  # 1x4x5x5
    """

    def __init__(self, window_size: int = 11, reduction: str = "none", max_val: float = 1.0) -> None:
        super(SSIM, self).__init__()
        self.window_size: int = window_size
        self.max_val: float = max_val
        self.reduction: str = reduction

        self.window: torch.Tensor = get_gaussian_kernel2d(
            (window_size, window_size), (1.5, 1.5))
        self.window = self.window.requires_grad_(False)  # need to disable gradients

        self.padding: int = _compute_zero_padding(window_size)

        self.C1: float = (0.01 * self.max_val) ** 2
        self.C2: float = (0.03 * self.max_val) ** 2

    def forward(self, img1: torch.Tensor, img2: torch.Tensor) -> torch.Tensor:
        if not torch.is_tensor(img1):
            raise TypeError(f'Input img1 type is not a torch.Tensor. Got {type(img1)}')
        if not torch.is_tensor(img2):
            raise TypeError(f'Input img2 type is not a torch.Tensor. Got {type(img2)}')
        if not img1.dim() == 4:
            raise ValueError(f'Invalid img1 shape, we expect BxCxHxW. Got: {img1.shape}')
        if not img2.dim() == 4:
            raise ValueError(f'Invalid img2 shape, we expect BxCxHxW. Got: {img2.shape}')
        if not img1.shape == img2.shape:
            raise ValueError(f'img1 and img2 shapes must be the same. Got: {img1.shape} and {img2.shape}')
        if not img1.device == img2.device:
            raise ValueError(f'img1 and img2 must be in the same device. Got: {img1.device} and {img2.device}')
        if not img1.dtype == img2.dtype:
            raise ValueError(f'img1 and img2 must be in the same dtype. Got: {img1.dtype} and {img2.dtype}')
        # prepare kernel
        b, c, h, w = img1.shape
        tmp_kernel: torch.Tensor = self.window.to(img1.device).to(img1.dtype)
        tmp_kernel = torch.unsqueeze(tmp_kernel, dim=0)
        # compute local mean per channel
        mu1: torch.Tensor = filter2D(img1, tmp_kernel)
        mu2: torch.Tensor = filter2D(img2, tmp_kernel)
        mu1_sq = mu1.pow(2)
        mu2_sq = mu2.pow(2)
        mu1_mu2 = mu1 * mu2
        # compute local sigma per channel
        sigma1_sq = filter2D(img1 * img1, tmp_kernel) - mu1_sq
        sigma2_sq = filter2D(img2 * img2, tmp_kernel) - mu2_sq
        sigma12 = filter2D(img1 * img2, tmp_kernel) - mu1_mu2
        ssim_map = ((2. * mu1_mu2 + self.C1) * (2. * sigma12 + self.C2)) / \
            ((mu1_sq + mu2_sq + self.C1) * (sigma1_sq + sigma2_sq + self.C2))
        loss = torch.clamp(-ssim_map + 1., min=0, max=1) / 2.
        if self.reduction == "mean":
            loss = torch.mean(loss)
        elif self.reduction == "sum":
            loss = torch.sum(loss)
        elif self.reduction == "none":
            pass
        return loss


######################
# functional interface
######################


def ssim(img1: torch.Tensor, img2: torch.Tensor,
         window_size: int, reduction: str = "none", max_val: float = 1.0) -> torch.Tensor:
    r"""Function that measures the Structural Similarity (SSIM) index between
    each element in the input `x` and target `y`.

    See :class:`~kornia.losses.SSIM` for details.
    """
    return SSIM(window_size, reduction, max_val)(img1, img2)


def normalize_kernel2d(input: torch.Tensor) -> torch.Tensor:
    r"""Normalizes both derivative and smoothing kernel.
    """
    if len(input.size()) < 2:
        raise TypeError(f'input should be at least 2D tensor. Got {input.size()}')
    norm: torch.Tensor = input.abs().sum(dim=-1).sum(dim=-1)
    return input / (norm.unsqueeze(-1).unsqueeze(-1))


def gaussian(window_size: int, sigma: float):
    x = torch.arange(window_size).float() - window_size // 2
    if window_size % 2 == 0:
        x = x + 0.5
    gauss = torch.exp((-x.pow(2.0) / float(2 * sigma ** 2)))
    return gauss / gauss.sum()


def get_gaussian_kernel1d(kernel_size: int, sigma: float, force_even: bool = False) -> torch.Tensor:
    r"""Function that returns Gaussian filter coefficients.

    Args:
        kernel_size (int): filter size. It should be odd and positive.
        sigma (float): gaussian standard deviation.
        force_even (bool): overrides requirement for odd kernel size.

    Returns:
        Tensor: 1D tensor with gaussian filter coefficients.

    Shape:
        - Output: :math:`(\text{kernel_size})`

    # Examples::

    #     >>> kornia.image.get_gaussian_kernel(3, 2.5)
    #     tensor([0.3243, 0.3513, 0.3243])

    #     >>> kornia.image.get_gaussian_kernel(5, 1.5)
    #     tensor([0.1201, 0.2339, 0.2921, 0.2339, 0.1201])
    """
    if (not isinstance(kernel_size, int) or (
            (kernel_size % 2 == 0) and not force_even) or (
            kernel_size <= 0)):
        raise TypeError(
            "kernel_size must be an odd positive integer. "
            f"Got {kernel_size}"
        )
    window_1d: torch.Tensor = gaussian(kernel_size, sigma)
    return window_1d


def get_gaussian_kernel2d(kernel_size: tuple[int, int], sigma: tuple[float, float],
                          force_even: bool = False) -> torch.Tensor:
    r"""Function that returns Gaussian filter matrix coefficients.

    Args:
        kernel_size (tuple[int, int]): filter sizes in the x and y direction.
         Sizes should be odd and positive.
        sigma (tuple[int, int]): gaussian standard deviation in the x and y
         direction.
        force_even (bool): overrides requirement for odd kernel size.

    Returns:
        Tensor: 2D tensor with gaussian filter matrix coefficients.

    Shape:
        - Output: :math:`(\text{kernel_size}_x, \text{kernel_size}_y)`

    # Examples::

    #     >>> kornia.image.get_gaussian_kernel2d((3, 3), (1.5, 1.5))
    #     tensor([[0.0947, 0.1183, 0.0947],
    #             [0.1183, 0.1478, 0.1183],
    #             [0.0947, 0.1183, 0.0947]])

    #     >>> kornia.image.get_gaussian_kernel2d((3, 5), (1.5, 1.5))
    #     tensor([[0.0370, 0.0720, 0.0899, 0.0720, 0.0370],
    #             [0.0462, 0.0899, 0.1123, 0.0899, 0.0462],
    #             [0.0370, 0.0720, 0.0899, 0.0720, 0.0370]])
    """
    if not isinstance(kernel_size, tuple) or len(kernel_size) != 2:
        raise TypeError(f"kernel_size must be a tuple of length two. Got {kernel_size}")
    if not isinstance(sigma, tuple) or len(sigma) != 2:
        raise TypeError(f"sigma must be a tuple of length two. Got {sigma}")
    ksize_x, ksize_y = kernel_size
    sigma_x, sigma_y = sigma
    kernel_x: torch.Tensor = get_gaussian_kernel1d(ksize_x, sigma_x, force_even)
    kernel_y: torch.Tensor = get_gaussian_kernel1d(ksize_y, sigma_y, force_even)
    kernel_2d: torch.Tensor = torch.matmul(
        kernel_x.unsqueeze(-1), kernel_y.unsqueeze(-1).t()
    )
    return kernel_2d


def compute_padding(kernel_size: tuple[int, int]) -> list[int]:
    """Computes padding tuple."""
    # 4 ints:  (padding_left, padding_right,padding_top,padding_bottom)
    # https://pytorch.org/docs/stable/nn.html#torch.nn.functional.pad
    assert len(kernel_size) == 2, kernel_size
    computed = [k // 2 for k in kernel_size]

    # for even kernels we need to do asymetric padding :(
    return [computed[1] - 1 if kernel_size[0] % 2 == 0 else computed[1],
            computed[1],
            computed[0] - 1 if kernel_size[1] % 2 == 0 else computed[0],
            computed[0]]


def filter2D(input: torch.Tensor, kernel: torch.Tensor,
             border_type: str = 'reflect',
             normalized: bool = False) -> torch.Tensor:
    r"""Function that convolves a tensor with a kernel.

    The function applies a given kernel to a tensor. The kernel is applied
    independently at each depth channel of the tensor. Before applying the
    kernel, the function applies padding according to the specified mode so
    that the output remains in the same shape.

    Args:
        input (torch.Tensor): the input tensor with shape of
          :math:`(B, C, H, W)`.
        kernel (torch.Tensor): the kernel to be convolved with the input
          tensor. The kernel shape must be :math:`(1, kH, kW)`.
        border_type (str): the padding mode to be applied before convolving.
          The expected modes are: ``'constant'``, ``'reflect'``,
          ``'replicate'`` or ``'circular'``. Default: ``'reflect'``.
        normalized (bool): If True, kernel will be L1 normalized.

    Return:
        torch.Tensor: the convolved tensor of same size and numbers of channels
        as the input.
    """
    if not isinstance(input, torch.Tensor):
        raise TypeError(f'Input type is not a torch.Tensor. Got {type(input)}')
    if not isinstance(kernel, torch.Tensor):
        raise TypeError(f'Input kernel type is not a torch.Tensor. Got {type(kernel)}')
    if not isinstance(border_type, str):
        raise TypeError(f'Input border_type is not string. Got {type(kernel)}')
    if not input.dim() == 4:
        raise ValueError(f'Invalid input shape, we expect BxCxHxW. Got: {input.shape}')
    if not kernel.dim() == 3:
        raise ValueError(f'Invalid kernel shape, we expect 1xHxW. Got: {kernel.shape}')
    borders_list: list[str] = ['constant', 'reflect', 'replicate', 'circular']
    if border_type not in borders_list:
        raise ValueError(f"Invalid border_type, we expect the following: {borders_list}."
                         f"Got: {border_type}")
    # prepare kernel
    b, c, h, w = input.shape
    tmp_kernel: torch.Tensor = kernel.unsqueeze(0).to(input.device).to(input.dtype)
    if normalized:
        tmp_kernel = normalize_kernel2d(tmp_kernel)
    # pad the input tensor
    height, width = tmp_kernel.shape[-2:]
    padding_shape: list[int] = compute_padding((height, width))
    input_pad: torch.Tensor = F.pad(input, padding_shape, mode=border_type)
    b, c, hp, wp = input_pad.shape
    # convolve the tensor with the kernel. Pick the fastest alg
    kernel_numel: int = height * width
    if kernel_numel > 81:
        return F.conv2d(input_pad.reshape(b * c, 1, hp, wp), tmp_kernel, padding=0, stride=1).view(b, c, h, w)
    return F.conv2d(input_pad, tmp_kernel.expand(c, -1, -1, -1), groups=c, padding=0, stride=1)



#!/usr/bin/env python3

from autovul.base.datasets import Dataset
from autovul.vision.environ import env
from autovul.vision.utils.data import Cutout

import torch
import torchvision.transforms as transforms
import argparse
import os

from typing import TYPE_CHECKING
from torchvision.datasets import VisionDataset  # TODO: python 3.10
import PIL.Image as Image
from typing import Union
if TYPE_CHECKING:
    import torch.utils.data


class ImageSet(Dataset):

    name: str = 'imageset'
    data_type: str = 'image'
    num_classes = 1000
    data_shape = [3, 224, 224]

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--dataset_normalize', dest='normalize',
                           action='store_true', help='use transforms.Normalize in dataset transform. '
                           '(It\'s used in model as the first layer by default.)')
        group.add_argument('--transform', choices=[None, 'bit', 'pytorch'])
        group.add_argument('--auto_augment', action='store_true', help='use auto augment')
        group.add_argument('--cutout', action='store_true', help='use cutout')
        group.add_argument('--cutout_length', type=int, help='cutout length')
        return group

    def __init__(self, norm_par: dict[str, list[float]] = None,
                 default_model: str = 'resnet18_comp',
                 normalize: bool = False, transform: str = None, auto_augment: bool = False,
                 cutout: bool = False, cutout_length: int = None, **kwargs):
        self.norm_par: dict[str, list[float]] = norm_par
        self.normalize = normalize
        self.transform = transform
        self.auto_augment = auto_augment
        self.cutout = cutout
        self.cutout_length = cutout_length
        super().__init__(default_model=default_model, **kwargs)
        self.param_list['imageset'] = ['data_shape', 'norm_par', 'normalize', 'transform', 'auto_augment', 'cutout']
        if cutout:
            self.param_list['imageset'].append('cutout_length')

    def get_transform(self, mode: str, normalize: bool = None) -> transforms.Compose:
        normalize = normalize if normalize is not None else self.normalize
        if self.transform == 'bit':
            return get_transform_bit(mode, self.data_shape)
        elif self.data_shape == [3, 224, 224]:
            transform = get_transform_imagenet(mode, use_tuple=self.transform != 'pytorch',
                                               auto_augment=self.auto_augment)
        elif self.data_shape in ([3, 16, 16], [3, 32, 32]):
            transform = get_transform_cifar(mode, auto_augment=self.auto_augment,
                                            cutout=self.cutout, cutout_length=self.cutout_length,
                                            data_shape=self.data_shape)
        else:
            transform = transforms.Compose([transforms.ToTensor()])
        if normalize and self.norm_par is not None:
            transform.transforms.append(transforms.Normalize(mean=self.norm_par['mean'], std=self.norm_par['std']))
        return transform

    def get_dataloader(self, mode: str = None, dataset: Dataset = None, batch_size: int = None, shuffle: bool = None,
                       num_workers: int = None, pin_memory=True, drop_last=False, **kwargs) -> torch.utils.data.DataLoader:
        if batch_size is None:
            batch_size = self.test_batch_size if mode == 'test' else self.batch_size
        if shuffle is None:
            shuffle = True if mode == 'train' else False
        num_workers = num_workers if num_workers is not None else self.num_workers
        if dataset is None:
            dataset = self.get_dataset(mode, **kwargs)
        if env['num_gpus'] == 0:
            pin_memory = False
        return torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=shuffle,
                                           num_workers=num_workers, pin_memory=pin_memory, drop_last=drop_last)

    @staticmethod
    def get_data(data: tuple[torch.Tensor, torch.Tensor], **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        return data[0].to(env['device'], non_blocking=True), data[1].to(env['device'], dtype=torch.long, non_blocking=True)

    def get_class_to_idx(self, **kwargs) -> dict[str, int]:
        if hasattr(self, 'class_to_idx'):
            return getattr(self, 'class_to_idx')
        return {str(i): i for i in range(self.num_classes)}

    def make_folder(self, img_type: str = '.png', **kwargs):
        mode_list: list[str] = ['train', 'valid'] if self.valid_set else ['train']
        class_to_idx = self.get_class_to_idx(**kwargs)
        idx_to_class = {v: k for k, v in class_to_idx.items()}
        for mode in mode_list:
            dataset: VisionDataset = self.get_org_dataset(mode, transform=None)
            class_counters = [0] * self.num_classes
            for image, target_class in list(dataset):
                image: Image.Image
                target_class: int
                class_name = idx_to_class[target_class]
                _dir = os.path.join(self.folder_path, self.name, mode, class_name)
                if not os.path.exists(_dir):
                    os.makedirs(_dir)
                image.save(os.path.join(_dir, f'{class_counters[target_class]}{img_type}'))
                class_counters[target_class] += 1


def get_transform_bit(mode: str, data_shape: list[int]) -> transforms.Compose:
    hyperrule = data_shape[-2] * data_shape[-1] < 96 * 96
    precrop, crop = (160, 128) if hyperrule else (512, 480)
    if mode == 'train':
        transform = transforms.Compose([
            transforms.Resize((precrop, precrop)),
            transforms.RandomCrop((crop, crop)),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ])
    else:
        transform = transforms.Compose([
            transforms.Resize((crop, crop)),
            transforms.ToTensor()])
    return transform


def get_transform_imagenet(mode: str, use_tuple: bool = False, auto_augment: bool = False) -> transforms.Compose:
    if mode == 'train':
        transform_list = [
            transforms.RandomResizedCrop((224, 224) if use_tuple else 224),
            transforms.RandomHorizontalFlip(),
            # transforms.ColorJitter(brightness=0.4, contrast=0.4, saturation=0.4),
        ]
        if auto_augment:
            transform_list.append(transforms.AutoAugment(transforms.AutoAugmentPolicy.IMAGENET))
        transform_list.append(transforms.ToTensor())
        transform = transforms.Compose(transform_list)
    else:
        transform = transforms.Compose([
            transforms.Resize((256, 256) if use_tuple else 256),
            transforms.CenterCrop((224, 224) if use_tuple else 224),
            transforms.ToTensor()])
    return transform


def get_transform_cifar(mode: str, auto_augment: bool = False,
                        cutout: bool = False, cutout_length: int = None,
                        data_shape: list[int] = [3, 32, 32]) -> transforms.Compose:
    if mode != 'train':
        return transforms.Compose([transforms.ToTensor()])
    cutout_length = data_shape[-1] // 2 if cutout_length is None else cutout_length
    transform_list = [
        transforms.RandomCrop(data_shape[-2:], padding=data_shape[-1] // 8),
        transforms.RandomHorizontalFlip(),
    ]
    if auto_augment:
        transform_list.append(transforms.AutoAugment(transforms.AutoAugmentPolicy.CIFAR10))
    transform_list.append(transforms.ToTensor())
    if cutout:
        transform_list.append(Cutout(cutout_length))
    return transforms.Compose(transform_list)



#!/usr/bin/env python3

# https://github.com/google-research/big_transfer

from autovul.vision.utils.model_archs.bit import KNOWN_MODELS, tf2th, StdConv2d, conv3x3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel
from autovul.vision.datasets import ImageNet

import torch
import torch.nn as nn
import torch.hub
import numpy as np
import os
from collections import OrderedDict


class _BiT(_ImageModel):
    def __init__(self, name: str = 'bit-m-r50x1', **kwargs):
        super().__init__(**kwargs)
        model_name = name.split('_')[0].upper().replace('BIT', 'BiT').replace('X', 'x')
        _model = KNOWN_MODELS[model_name](head_size=1)
        self.features = nn.Sequential()
        if 'comp' in name:
            assert 'official' not in name, name
            conv: StdConv2d = _model.root.conv
            self.features.add_module('conv', conv3x3(conv.in_channels, conv.out_channels))
        else:
            self.features.add_module('conv', _model.root.conv)
            if 'official' not in name:
                pool: nn.MaxPool2d = _model.root.pool
                _model.root.pool = nn.MaxPool2d(pool.kernel_size, pool.stride, padding=1)
            else:
                self.features.add_module('pad', _model.root.pad)
            self.features.add_module('pool', _model.root.pool)
        for name, child in _model.body.named_children():
            self.features.add_module(name, child)
        self.features.add_module('gn', _model.head.gn)
        self.features.add_module('relu', _model.head.relu)
        fc: nn.Conv2d = _model.head.conv
        self.classifier = self.define_classifier(conv_dim=fc.in_channels,
                                                 num_classes=self.num_classes, fc_depth=1)


class BiT(ImageModel):
    available_models = ['bit', 'bit_comp', 'bit_official',
                        'bit-m-r50x1', 'bit-m-r50x3', 'bit-m-r101x1', 'bit-m-r101x3', 'bit-m-r152x2', 'bit-m-r152x4',
                        'bit-s-r50x1', 'bit-s-r50x3', 'bit-s-r101x1', 'bit-s-r101x3', 'bit-s-r152x2', 'bit-s-r152x4',
                        'bit-m-r50x1_comp', 'bit-m-r50x3_comp', 'bit-m-r101x1_comp',
                        'bit-m-r101x3_comp', 'bit-m-r152x2_comp', 'bit-m-r152x4_comp',
                        'bit-s-r50x1_comp', 'bit-s-r50x3_comp', 'bit-s-r101x1_comp',
                        'bit-s-r101x3_comp', 'bit-s-r152x2_comp', 'bit-s-r152x4_comp',
                        'bit-m-r50x1_official', 'bit-m-r50x3_official', 'bit-m-r101x1_official',
                        'bit-m-r101x3_official', 'bit-m-r152x2_official', 'bit-m-r152x4_official',
                        'bit-s-r50x1_official', 'bit-s-r50x3_official', 'bit-s-r101x1_official',
                        'bit-s-r101x3_official', 'bit-s-r152x2_official', 'bit-s-r152x4_official']

    def __init__(self, name: str = 'bit',
                 pretrained_dataset: str = 'm', layer: int = 50, width_factor: int = 1,
                 model: type[_BiT] = _BiT, norm_par: dict[str, list[float]] = None,
                 official: bool = False, **kwargs):
        name = self.parse_name(name, pretrained_dataset, layer, width_factor)
        if official and norm_par is None:
            norm_par = {'mean': [0.5, 0.5, 0.5],
                        'std': [0.5, 0.5, 0.5], }
        super().__init__(name=name, width_factor=width_factor,
                         model=model, norm_par=norm_par,
                         official=official, **kwargs)

    @staticmethod
    def parse_name(name: str, pretrained_dataset: str = 'm', layer: int = 50, width_factor: int = 1) -> str:
        full_name_list = name.split('_')
        name_list = full_name_list[0].lower().split('-')
        assert name_list[0] == 'bit', name
        if len(name_list) != 1:
            for element in name_list[1:]:
                if element[0] == 'r':
                    sub_list = element[1:].split('x')
                    layer = int(sub_list[0])
                    if len(sub_list) == 2:
                        width_factor = int(sub_list[1])
                else:
                    assert len(element) == 1, name
                    pretrained_dataset = element
        full_name_list[0] = '-'.join(['bit', pretrained_dataset, f'r{layer:d}x{width_factor:d}'])
        return '_'.join(full_name_list)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        # TODO: map_location argument
        assert 'official' in self.name and 'comp' not in self.name, self.name
        model_name = self.name.split('_')[0].upper().replace('BIT', 'BiT').replace('X', 'x')
        if isinstance(self.dataset, ImageNet):
            model_name += '-ILSVRC2012'
        url = f'https://storage.googleapis.com/bit_models/{model_name}.npz'
        print('get official model weights from: ', url)
        file_path = os.path.join(torch.hub.get_dir(), 'bit', f'{model_name}.npz')
        if not os.path.exists(file_path):
            torch.hub.download_url_to_file(url, file_path)
        weights: dict[str, np.ndarray] = np.load(file_path)
        _dict = OrderedDict()
        _dict['features.conv.weight'] = tf2th(weights['resnet/root_block/standardized_conv2d/kernel'])
        for block_num in range(4):
            block_name = f'block{block_num+1:d}'
            block: nn.Sequential = getattr(self._model.features, block_name)
            for unit_name, unit in block.named_children():
                prefix = '/'.join(['resnet', block_name, unit_name, ''])
                dict_prefix = '.'.join(['features', block_name, unit_name, ''])
                _dict[dict_prefix + 'gn1.weight'] = tf2th(weights[prefix + 'a/group_norm/gamma'])
                _dict[dict_prefix + 'gn1.bias'] = tf2th(weights[prefix + 'a/group_norm/beta'])
                _dict[dict_prefix + 'conv1.weight'] = tf2th(weights[prefix + 'a/standardized_conv2d/kernel'])
                _dict[dict_prefix + 'gn2.weight'] = tf2th(weights[prefix + 'b/group_norm/gamma'])
                _dict[dict_prefix + 'gn2.bias'] = tf2th(weights[prefix + 'b/group_norm/beta'])
                _dict[dict_prefix + 'conv2.weight'] = tf2th(weights[prefix + 'b/standardized_conv2d/kernel'])
                _dict[dict_prefix + 'gn3.weight'] = tf2th(weights[prefix + 'c/group_norm/gamma'])
                _dict[dict_prefix + 'gn3.bias'] = tf2th(weights[prefix + 'c/group_norm/beta'])
                _dict[dict_prefix + 'conv3.weight'] = tf2th(weights[prefix + 'c/standardized_conv2d/kernel'])
                if hasattr(unit, 'downsample'):
                    weight = tf2th(weights[prefix + 'a/proj/standardized_conv2d/kernel'])
                    _dict[dict_prefix + 'downsample.weight'] = weight
        _dict['features.gn.weight'] = tf2th(weights['resnet/group_norm/gamma'])
        _dict['features.gn.bias'] = tf2th(weights['resnet/group_norm/beta'])
        _dict['classifier.fc.weight'] = tf2th(weights['resnet/head/conv2d/kernel']).flatten(1)
        _dict['classifier.fc.bias'] = tf2th(weights['resnet/head/conv2d/bias'])
        return _dict



#!/usr/bin/env python3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel

import torchvision.models
from torchvision.models.mnasnet import _MODEL_URLS as urls
import re

import torch
from collections import OrderedDict


class _MNASNet(_ImageModel):

    def __init__(self, mnas_alpha: float, **kwargs):
        super().__init__(**kwargs)
        _model = torchvision.models.MNASNet(mnas_alpha, num_classes=self.num_classes)
        self.features = _model.layers
        self.classifier = _model.classifier
        # conv: nn.Conv2d = self.features[0]
        # self.features[0] = nn.Conv2d(3, conv.out_channels, 3, padding=1, stride=1, bias=False)


class MNASNet(ImageModel):
    available_models = ['mnasnet', 'mnasnet0_5', 'mnasnet0_75', 'mnasnet1_0', 'mnasnet1_3']
    model_urls = urls

    def __init__(self, name: str = 'mnasnet', mnas_alpha: float = 1.0,
                 model: type[_MNASNet] = _MNASNet, **kwargs):
        name, self.mnas_alpha = self.parse_name(name, mnas_alpha)
        super().__init__(name=name, mnas_alpha=self.mnas_alpha, model=model, **kwargs)

    @staticmethod
    def parse_name(name: str, mnas_alpha: float = 1.0) -> tuple[str, float]:
        name_list: list[str] = re.findall('[a-zA-Z]+|[\d_.]+', name)
        name = name_list[0]
        if len(name_list) > 1:
            assert len(name_list) == 2
            mnas_alpha = float(name_list[1].replace('_', '.'))
        return f'{name}{mnas_alpha:.1f}'.replace('.', '_'), mnas_alpha

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        url = self.model_urls[self.parse_name('mnasnet', self.mnas_alpha)[0]]
        _dict = super().get_official_weights(url=url)
        new_dict = OrderedDict()
        for key, value in _dict.items():
            if key.startswith('layers.'):
                key = 'features.' + key[7:]
            new_dict[key] = value
        return new_dict



#!/usr/bin/env python3
from autovul.vision.datasets import ImageSet
from autovul.vision.models.imagemodel import _ImageModel, ImageModel
from autovul.vision.utils.model_archs.darts import FeatureExtractor, AuxiliaryHead, Genotype
from autovul.vision.utils.model_archs.darts import genotypes

import torch
import torch.nn as nn
from torchvision.datasets.utils import download_file_from_google_drive
import os
from collections import OrderedDict

from typing import TYPE_CHECKING
import argparse  # TODO: python 3.10
if TYPE_CHECKING:
    pass

url = {
    'cifar10': '1Y13i4zKGKgjtWBdC0HWLavjO7wvEiGOc',
    'ptb': '1Mt_o6fZOlG-VDF3Q5ModgnAJ9W6f_av2',
    'imagenet': '1AKr6Y_PoYj7j0Upggyzc26W0RVdg4CVX'
}


class _DARTS(_ImageModel):
    def __init__(self, auxiliary: bool = False, **kwargs):
        super().__init__(**kwargs)
        self.features: FeatureExtractor
        self.classifier = self.define_classifier(conv_dim=self.features.feats_dim,
                                                 num_classes=self.num_classes, fc_depth=1)
        self.auxiliary_head: nn.Sequential = None
        if auxiliary:
            self.auxiliary_head = AuxiliaryHead(C=self.features.aux_dim, num_classes=self.num_classes)

    @staticmethod
    def define_features(genotype: Genotype = genotypes.darts,
                        C: int = 36, layers: int = 20,
                        dropout_p: float = 0.2, **kwargs) -> FeatureExtractor:
        return FeatureExtractor(genotype, C, layers, dropout_p, **kwargs)


class DARTS(ImageModel):
    available_models = ['darts']

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--model_arch', help='Model Architecture (genotype name), defaults to be "darts"')
        group.add_argument('--auxiliary', action='store_true', help='enable auxiliary classifier during training.')
        group.add_argument('--auxiliary_weight', type=float, help='weight for auxiliary loss, defaults to be 0.4')
        return group

    def __init__(self, name: str = 'darts', model_arch: str = 'darts',
                 layers: int = 20, C: int = 36, dropout_p: float = 0.2,
                 auxiliary: bool = False, auxiliary_weight: float = 0.4,
                 genotype: Genotype = None, model: type[_DARTS] = _DARTS, **kwargs):
        # TODO: ImageNet parameter settings
        if genotype is None:
            model_arch = model_arch.lower()
            name = model_arch
            try:
                genotype = getattr(genotypes, model_arch)
            except AttributeError as e:
                print('Available Model Architectures: ')
                model_arch_list = [element for element in dir(genotypes) if '__' not in element and
                                   element not in ['Genotype', 'PRIMITIVES', 'namedtuple']]
                print(model_arch_list)
                raise e
        self.layers = layers
        self.C = C
        self.dropout_p = dropout_p
        self.genotype = genotype
        self.auxiliary = auxiliary
        self.auxiliary_weight = auxiliary_weight
        super().__init__(name=name, layers=layers, C=C, dropout_p=dropout_p,
                         genotype=genotype, model=model,
                         auxiliary=auxiliary, **kwargs)
        self._model: _DARTS
        self.param_list['darts'] = ['layers', 'C', 'dropout_p', 'genotype']
        if auxiliary:
            self.param_list['darts'].insert(0, 'auxiliary_weight')

    def loss(self, _input: torch.Tensor = None, _label: torch.Tensor = None,
             _output: torch.Tensor = None, amp: bool = False, **kwargs) -> torch.Tensor:
        if self.auxiliary:
            assert isinstance(self._model.auxiliary_head, nn.Sequential)
            if amp:
                with torch.cuda.amp.autocast():
                    return self.loss_with_aux(_input, _label)
            return self.loss_with_aux(_input, _label)
        else:
            return super().loss(_input, _label, _output, **kwargs)

    def loss_with_aux(self, _input: torch.Tensor = None, _label: torch.Tensor = None) -> torch.Tensor:
        feats, feats_aux = self._model.features.forward_with_aux(self._model.normalize(_input))
        logits: torch.Tensor = self._model.classifier(self._model.flatten(self._model.pool(feats)))
        logits_aux: torch.Tensor = self._model.auxiliary_head(feats_aux)
        return super().loss(_output=logits, _label=_label) \
            + self.auxiliary_weight * super().loss(_output=logits_aux, _label=_label)

    def load(self, *args, strict: bool = False, **kwargs):
        return super().load(*args, strict=strict, **kwargs)

    def get_official_weights(self, dataset: str = None, **kwargs) -> OrderedDict[str, torch.Tensor]:
        assert str(self.genotype) == str(genotypes.darts)
        if dataset is None and isinstance(self.dataset, ImageSet):
            dataset = self.dataset.name
        file_name = f'darts_{dataset}.pt'
        folder_path = os.path.join(torch.hub.get_dir(), 'darts')
        download_file_from_google_drive(file_id=url[dataset], root=folder_path, filename=file_name)
        print('get official model weights from Google Drive: ', url[dataset])
        _dict: OrderedDict[str, torch.Tensor] = torch.load(os.path.join(self.folder_path, file_name),
                                                           map_location='cpu')
        if 'state_dict' in _dict.keys():
            _dict = _dict['state_dict']

        new_dict: OrderedDict[str, torch.Tensor] = self.state_dict()
        old_keys = list(_dict.keys())
        new_keys = list(new_dict.keys())
        new2old: dict[str, str] = {}
        i = 0
        j = 0
        while(i < len(new_keys) and j < len(old_keys)):
            if 'num_batches_tracked' in new_keys[i]:
                i += 1
                continue
            if 'auxiliary_head' not in new_keys[i] and 'auxiliary_head' in old_keys[j]:
                j += 1
                continue
            new2old[new_keys[i]] = old_keys[j]
            i += 1
            j += 1
        for i, key in enumerate(new_keys):
            if 'num_batches_tracked' in key:
                new_dict[key] = torch.tensor(0)
            else:
                new_dict[key] = _dict[new2old[key]]
        return new_dict



#!/usr/bin/env python3

from autovul.vision.models.imagemodel import ImageModel

from .natsbench import NATSbench

from .darts import DARTS
from .enas import ENAS
from .lanet import LaNet
from .mnasnet import MNASNet
from .pnasnet import PNASNet
from .proxylessnas import ProxylessNAS

__all__ = ['NATSbench', 'DARTS', 'ENAS', 'LaNet', 'MNASNet', 'PNASNet', 'ProxylessNAS']

class_dict: dict[str, type[ImageModel]] = {
    'natsbench': NATSbench,
    'darts': DARTS,
    'enas': ENAS,
    'lanet': LaNet,
    'mnasnet': MNASNet,
    'pnasnet': PNASNet,
    'proxylessnas': ProxylessNAS,
}



#!/usr/bin/env python3

# python scripts/membership.py --model densenet121_comp --dataset cifar100 --pretrain

import autovul.vision
from autovul.vision import to_numpy

from scipy.special import softmax
import argparse
import warnings
import numpy as np

warnings.filterwarnings("ignore")

if __name__ == '__main__':

    parser = argparse.ArgumentParser()
    autovul.vision.environ.add_argument(parser)
    autovul.vision.datasets.add_argument(parser)
    autovul.vision.models.add_argument(parser)
    args = parser.parse_args()
    env = autovul.vision.environ.create(**args.__dict__)
    dataset = autovul.vision.datasets.create(**args.__dict__)
    model = autovul.vision.models.create(dataset=dataset, **args.__dict__)

    import torch
    from sklearn import metrics
    from trojanzoo.utils.data import dataset_to_list
    from art.estimators.classification import PyTorchClassifier  # type: ignore
    classifier = PyTorchClassifier(
        model=model._model,
        loss=model.criterion,
        input_shape=dataset.data_shape,
        nb_classes=model.num_classes,
        clip_values=(0, 1)
    )
    max_iter = 10
    max_eval = 100
    sample_size = 50
    init_size = 50
    init_eval = 25

    from art.attacks.evasion.hop_skip_jump import HopSkipJump  # type: ignore

    x_train, y_train = dataset_to_list(dataset.get_dataset('train'))
    x_train, y_train = to_numpy(torch.stack(x_train)), to_numpy(y_train)
    t_idx = np.arange(len(x_train))

    x_valid, y_valid = dataset_to_list(dataset.get_dataset('valid'))
    x_valid, y_valid = to_numpy(torch.stack(x_valid)), to_numpy(y_valid)
    preds = np.amax(softmax(classifier.predict(x_valid), axis=1), axis=1)
    v_idx = np.arange(len(x_valid))[(preds <= 0.99999) & (preds >= 0.999)]

    x_valid = x_valid[v_idx]
    y_valid = y_valid[v_idx]
    v_idx = np.arange(len(x_valid))

    np.random.seed(30)
    np.random.shuffle(t_idx)
    np.random.shuffle(v_idx)
    train_idx = t_idx[:sample_size]
    valid_idx = v_idx[:sample_size]

    x = np.concatenate((x_train[train_idx], x_valid[valid_idx]))
    y = np.concatenate((y_train[train_idx], y_valid[valid_idx]))
    y_truth = np.concatenate(([0] * sample_size, [1] * sample_size))

    hsj = HopSkipJump(classifier=classifier, targeted=False, norm=2, max_iter=max_iter,
                      init_size=init_size, init_eval=init_eval, max_eval=max_eval, verbose=False)
    x_adv = hsj.generate(x=x, y=None)
    distance = np.linalg.norm((x_adv - x).reshape((x.shape[0], -1)), ord=np.inf, axis=1)

    fpr, tpr, _ = metrics.roc_curve(y_truth, distance)
    auc = metrics.auc(fpr, tpr)
    auc = auc if auc >= 0.5 else 1 - auc
    print(f'{model.name:20}    AUC:  {str(auc)}')



#!/usr/bin/env python3

from autovul.vision.utils.model import Conv2d_SAME
from autovul.base.utils import to_tensor
from autovul.base.models import Model
from autovul.vision.datasets import ImageSet

import torch
import torch.nn as nn

from typing import Iterator, Tuple, Union
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler

# Note that MagNet requires "eval" mode to train.


class _MagNet(nn.Module):
    """docstring for Model"""

    def __init__(self, structure: list[Tuple[int, str]] = [3, 'average', 3],
                 activation: str = 'sigmoid', channel: int = 3, **kwargs):
        super().__init__()

        activation_fn = nn.ReLU()
        if activation == 'sigmoid':
            activation_fn = nn.Sigmoid()

        self.encoder = nn.Sequential()
        self.decoder = nn.Sequential()

        in_channels = channel
        for i, layer in enumerate(structure):
            if isinstance(layer, int):
                conv = Conv2d_SAME(in_channels=in_channels, out_channels=structure[i], kernel_size=(3, 3))
                in_channels = structure[i]
                bn = nn.BatchNorm2d(structure[i])
                self.encoder.add_module(f'conv{i+1:d}', conv)
                self.encoder.add_module(f'bn{i+1:d}', bn)
                self.encoder.add_module(f'{activation}{i+1:d}', activation_fn)
            else:
                assert isinstance(layer, str)
                module = nn.MaxPool2d(kernel_size=(2, 2)) if layer == 'max' else nn.AvgPool2d(kernel_size=(2, 2))
                self.encoder.add_module('pool', module)

        for i, layer in enumerate(reversed(structure)):
            if isinstance(layer, int):
                conv = Conv2d_SAME(in_channels=in_channels, out_channels=structure[i], kernel_size=(3, 3))
                in_channels = structure[i]
                bn = nn.BatchNorm2d(structure[i])
                self.decoder.add_module(f'conv{i+1:d}', conv)
                self.decoder.add_module(f'bn{i+1:d}', bn)
                self.decoder.add_module(f'{activation}{i+1:d}', activation_fn)
            else:
                assert isinstance(layer, str)
                self.decoder.add_module('pool', nn.Upsample(scale_factor=(2, 2)))
        conv = Conv2d_SAME(structure[0], channel, kernel_size=(3, 3))
        bn = nn.BatchNorm2d(channel)
        self.decoder.add_module('conv', conv)
        self.decoder.add_module('bn', bn)
        self.decoder.add_module('activation', activation_fn)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class MagNet(Model):
    available_models = ['magnet']

    def __init__(self, name: str = 'magnet', dataset: ImageSet = None, model: type = _MagNet,
                 structure: list = None, activation: str = None, v_noise: float = 0.1, **kwargs):
        self.v_noise: float = v_noise
        if structure is None:
            if dataset.data_shape[0] == 1:
                structure = [3, "average", 3]
            else:
                structure = [32]
        if activation is None:
            if dataset.data_shape[0] == 1:
                activation = 'sigmoid'
            else:
                activation = 'relu'
        super().__init__(name=name, dataset=dataset, model=model,
                         structure=structure, activation=activation,
                         channel=dataset.data_shape[0], **kwargs)

    def get_data(self, data: tuple[torch.Tensor], v_noise: float = None, mode='train'):
        if v_noise is None:
            v_noise = self.v_noise
        _input = data[0]
        if mode == 'train':
            noise: torch.Tensor = torch.normal(mean=0.0, std=v_noise, size=_input.shape)
            data[0] = (_input + noise).clamp(0.0, 1.0)
            data[1] = _input.detach()
        else:
            data[0] = _input.detach()
            data[1] = _input.clone().detach()
        return to_tensor(data[0]), to_tensor(data[1])

    def define_optimizer(self, parameters: Union[str, Iterator[nn.Parameter]] = 'full',
                         OptimType: Union[str, type[Optimizer]] = 'Adam',
                         lr: float = 0.1, momentum: float = 0.0, weight_decay: float = 1e-9,
                         lr_scheduler: bool = False, T_max: int = None,
                         **kwargs) -> tuple[Optimizer, _LRScheduler]:
        return super().define_optimizer(parameters=parameters, OptimType=OptimType, lr=lr,
                                        momentum=momentum, weight_decay=weight_decay,
                                        lr_scheduler=lr_scheduler, T_max=T_max, **kwargs)

    # define MSE loss function
    def define_criterion(self, **kwargs):
        entropy_fn = nn.MSELoss()

        def loss_fn(_output: torch.Tensor, _label: torch.Tensor):
            _output = _output.to(device=_label.device, dtype=_label.dtype)
            return entropy_fn(_output, _label)
        return loss_fn

    def accuracy(self, _output: torch.Tensor, _label: torch.Tensor, num_classes: int = None, topk=(1, 5)):
        res = []
        for k in topk:
            res.append(-self.criterion(_output, _label))
        return res



#!/usr/bin/env python3

from autovul.vision.models.imagemodel import ImageModel

from .alexnet import AlexNet
from .bit import BiT
from .densenet import DenseNet
from .dla import DLA
from .dpn import DPN
from .mobilenet import MobileNet
from .net import Net
from .resnet import ResNet
from .shufflenetv2 import ShuffleNetV2
from .vgg import VGG

__all__ = ['AlexNet', 'BiT', 'DenseNet', 'DLA', 'DPN',
           'MobileNet', 'Net', 'ResNet', 'ShuffleNetV2', 'VGG']

class_dict: dict[str, type[ImageModel]] = {
    'alexnet': AlexNet,
    'bit': BiT,
    'densenet': DenseNet,
    'dla': DLA,
    'dpn': DPN,
    'mobilenet': MobileNet,
    'net': Net,
    'resnet': ResNet,
    'resnext': ResNet,
    'shufflenetv2': ShuffleNetV2,
    'vgg': VGG,
}



#!/usr/bin/env python3
# https://github.com/D-X-Y/AutoDL-Projects/blob/main/xautodl/datasets/DownsampledImageNet.py

##################################################
# Copyright (c) Xuanyi Dong [GitHub D-X-Y], 2019 #
##################################################
import os
import pickle
from PIL import Image
import numpy as np

from torchvision.datasets.utils import check_integrity
from torchvision.datasets import VisionDataset

from collections.abc import Callable
from typing import Optional


class DownsampledImageNet(VisionDataset):
    # http://image-net.org/download-images
    # A Downsampled Variant of ImageNet as an Alternative to the CIFAR datasets
    # https://arxiv.org/pdf/1707.08819.pdf

    data_shape: list[int] = []
    train_list: list[list[str]] = []
    test_list: list[list[str]] = []

    def __init__(self, root: str, train: bool = True,
                 num_classes: int = None,
                 transform: Optional[Callable] = None,
                 target_transform: Optional[Callable] = None,
                 download: bool = False,
                 ) -> None:
        super().__init__(root, transform=transform,
                         target_transform=target_transform)

        self.num_classes = num_classes
        self.train = train  # training set or test set

        if download:
            self.download()

        if not self._check_integrity():
            raise RuntimeError('Dataset not found or corrupted.')

        if self.train:
            downloaded_list = self.train_list
        else:
            downloaded_list = self.test_list

        self.data = []
        self.targets: list[int] = []

        # now load the picked numpy arrays
        for file_name, checksum in downloaded_list:
            file_path = os.path.join(self.root, file_name)
            with open(file_path, 'rb') as f:
                entry = pickle.load(f, encoding='latin1')
                self.data.append(entry['data'])
                self.targets.extend([label - 1 for label in entry['labels']])
        self.data = np.vstack(self.data).reshape(-1, *self.data_shape)
        self.data = self.data.transpose((0, 2, 3, 1))  # convert to HWC

        if num_classes is not None:
            assert 0 < num_classes < 1000, f'{num_classes=}'
            new_data, new_targets = [], []
            for I, L in zip(self.data, self.targets):
                if L < num_classes:
                    new_data.append(I)
                    new_targets.append(L)
            self.data = new_data
            self.targets = new_targets

    def __repr__(self) -> str:
        lines = super().__repr__().split('\n')
        lines.insert(1, f'Number of Classes: {self.num_classes}')
        return '\n'.join(lines)

    def __getitem__(self, index: int) -> tuple[Image.Image, int]:
        img, target = self.data[index], self.targets[index]
        img = Image.fromarray(img)
        if self.transform is not None:
            img = self.transform(img)
        return img, target

    def __len__(self) -> int:
        return len(self.data)

    def _check_integrity(self) -> bool:
        root = self.root
        for filename, md5 in (self.train_list + self.test_list):
            fpath = os.path.join(root, filename)
            if not check_integrity(fpath, md5):
                return False
        return True

    def download(self) -> None:
        raise NotImplementedError()

    def extra_repr(self) -> str:
        return "Split: {}".format("Train" if self.train is True else "Test")


class ImageNet16(DownsampledImageNet):
    data_shape: list[int] = [3, 16, 16]
    train_list = [
        ['train_data_batch_1', '27846dcaa50de8e21a7d1a35f30f0e91'],
        ['train_data_batch_2', 'c7254a054e0e795c69120a5727050e3f'],
        ['train_data_batch_3', '4333d3df2e5ffb114b05d2ffc19b1e87'],
        ['train_data_batch_4', '1620cdf193304f4a92677b695d70d10f'],
        ['train_data_batch_5', '348b3c2fdbb3940c4e9e834affd3b18d'],
        ['train_data_batch_6', '6e765307c242a1b3d7d5ef9139b48945'],
        ['train_data_batch_7', '564926d8cbf8fc4818ba23d2faac7564'],
        ['train_data_batch_8', 'f4755871f718ccb653440b9dd0ebac66'],
        ['train_data_batch_9', 'bb6dd660c38c58552125b1a92f86b5d4'],
        ['train_data_batch_10', '8f03f34ac4b42271a294f91bf480f29b'],
    ]
    test_list = [
        ['val_data', '3410e3017fdaefba8d5073aaa65e4bd6'],
    ]


class ImageNet32(DownsampledImageNet):
    data_shape: list[int] = [3, 32, 32]
    train_list = [
        ['train_data_batch_1', 'dd6683a336ab645d336f7b47c67d8456'],
        ['train_data_batch_2', 'b9b1f5ad237638a41e80944fc03b42af'],
        ['train_data_batch_3', 'e33306a0f4234b5c5812a32845bfbd04'],
        ['train_data_batch_4', '4db6d81c460a5170a03ef03e1ddd21dd'],
        ['train_data_batch_5', 'baa0f1d06099f55e5fe50d43fca4d334'],
        ['train_data_batch_6', 'bba3bc6e83e0703df1ec7e277e0b6800'],
        ['train_data_batch_7', '48dd0fd4511b39d518dbabba2f5b36b7'],
        ['train_data_batch_8', '46583e44f926baaeacbf8c333dc6110f'],
        ['train_data_batch_9', '7c82c36ce028b457204f1080d1793669'],
        ['train_data_batch_10', '14e3a6ce19d584ba97208a1d9f551e63'],
    ]
    test_list = [
        ['val_data', '4836a1eec28cd4476eb017126cd0f059'],
    ]



#!/usr/bin/env python3
from torchvision.models.resnet import conv3x3
from autovul.vision.models.imagemodel import _ImageModel, ImageModel
from autovul.vision.utils.model_archs import dla

import torch.nn as nn

import torch
from collections import OrderedDict
from collections.abc import Callable

urls = {
    'dla34': 'http://dl.yf.io/dla/models/imagenet/dla34-ba72cf86.pth',
    'dla46_c': 'http://dl.yf.io/dla/models/imagenet/dla46_c-2bfd52c3.pth',
    'dla46x_c': 'http://dl.yf.io/dla/models/imagenet/dla46x_c-d761bae7.pth',
    'dla60x_c': 'http://dl.yf.io/dla/models/imagenet/dla60x_c-b870c45c.pth',
    'dla60': 'http://dl.yf.io/dla/models/imagenet/dla60-24839fc4.pth',
    'dla60x': 'http://dl.yf.io/dla/models/imagenet/dla60x-d15cacda.pth',
    'dla102': 'http://dl.yf.io/dla/models/imagenet/dla102-d94d9790.pth',
    'dla102x': 'http://dl.yf.io/dla/models/imagenet/dla102x-ad62be81.pth',
    'dla102x2': 'http://dl.yf.io/dla/models/imagenet/dla102x2-262837b6.pth',
    'dla169': 'http://dl.yf.io/dla/models/imagenet/dla169-0914e092.pth',
    # 'dla60_res2net':
    #     'https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-res2net/res2net_dla60_4s-d88db7f9.pth',
    # 'dla60_res2next':
    #     'https://github.com/rwightman/pytorch-image-models/releases/download/v0.1-res2net/res2next_dla60_4s-d327927b.pth',
}


class _DLA(_ImageModel):

    def __init__(self, name: str = 'dla', **kwargs):
        super().__init__(**kwargs)
        ModelClass: Callable[..., dla.DLA] = getattr(dla, name.replace('_comp', ''))
        if 'comp' in name:
            _model = ModelClass(num_classes=self.num_classes, strides=[1, 2, 2, 2])
            conv1: nn.Conv2d = _model.features[0][0]
            _model.features[0][0] = dla.conv3x3(conv1.in_channels, conv1.out_channels)  # stem.conv kernel_size: 7 -> 3
            conv3: nn.Conv2d = _model.features[2][0]
            _model.features[2][0] = dla.conv3x3(conv3.in_channels, conv3.out_channels)  # layer2.conv stride: 2 -> 1
        else:
            _model = ModelClass(num_classes=self.num_classes)
        self.features = _model.features
        self.classifier = _model.classifier


class DLA(ImageModel):
    available_models = ['dla', 'dla_comp',
                        'dla34', 'dla46_c', 'dla46x_c', 'dla60x_c', 'dla60', 'dla60x',
                        'dla102', 'dla102x', 'dla102x2', 'dla169',
                        'dla34_comp', 'dla46_c_comp', 'dla46x_c_comp', 'dla60x_c_comp', 'dla60_comp', 'dla60x_comp',
                        'dla102_comp', 'dla102x_comp', 'dla102x2_comp', 'dla169_comp']
    model_urls = urls

    def __init__(self, name: str = 'dla', layer: int = 34, model: type[_DLA] = _DLA, **kwargs):
        super().__init__(name=name, layer=layer, model=model, **kwargs)

    def get_official_weights(self, **kwargs) -> OrderedDict[str, torch.Tensor]:
        old_dict = super().get_official_weights(**kwargs)
        new_dict: OrderedDict[str, torch.Tensor] = self.state_dict()
        old_keys = list(old_dict.keys())
        new_keys = list(new_dict.keys())
        new2old: dict[str, str] = {}
        i = 0
        j = 0
        while(i < len(new_keys) and j < len(old_keys)):
            if 'num_batches_tracked' in new_keys[i]:
                i += 1
                continue
            if 'project' not in new_keys[i] and 'project' in old_keys[j]:
                j += 1
                continue
            # print(f'{new_keys[i]:60} {old_keys[j]}')
            new2old[new_keys[i]] = old_keys[j]
            i += 1
            j += 1
        for i, key in enumerate(new_keys):
            if 'num_batches_tracked' in key:
                new_dict[key] = torch.tensor(0)
            elif 'fc.weight' in key:
                weight = old_dict[new2old[key]]
                new_dict[key] = weight.flatten(1)
            else:
                new_dict[key] = old_dict[new2old[key]]
        return new_dict



#!/usr/bin/env python3

from autovul.vision.shortcut.pgd import PGD
from autovul.vision.datasets import ImageSet
from autovul.vision.utils import apply_cmap
from autovul.base.models import _Model, Model
from autovul.base.environ import env
from autovul.base.utils import add_noise
from autovul.base.utils.fim import KFAC, EKFAC

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision.transforms import Normalize
import re

from typing import TYPE_CHECKING
from typing import Union
from torch.optim.optimizer import Optimizer
from torch.optim.lr_scheduler import _LRScheduler
import argparse
from matplotlib.colors import Colormap  # type: ignore  # TODO
from collections.abc import Callable
if TYPE_CHECKING:
    import torch.utils.data

from matplotlib.cm import get_cmap  # type: ignore  # TODO
jet = get_cmap('jet')

log_softmax = nn.LogSoftmax(1)
criterion_kl = nn.KLDivLoss(reduction='batchmean')


def replace_bn_to_gn(model: nn.Module) -> None:
    for name, module in model.named_children():
        replace_bn_to_gn(module)
        if isinstance(module, nn.BatchNorm2d):
            device = module.weight.device
            gn = nn.GroupNorm(module.num_features, module.num_features, device=device)
            setattr(model, name, gn)


def set_first_layer_channel(model: nn.Module, channel: int = 3, **kwargs) -> None:
    for name, module in model.named_children():
        if len(list(module.children())):
            set_first_layer_channel(module, channel=channel)
        elif isinstance(module, nn.Conv2d):
            if module.in_channels == channel:
                return
            keys = ['out_channels', 'kernel_size', 'stride', 'padding']
            args = {key: getattr(module, key) for key in keys}
            args['device'] = module.weight.device
            args.update(kwargs)
            new_conv = nn.Conv2d(in_channels=channel, bias=False, **args)
            setattr(model, name, new_conv)
        break


class _ImageModel(_Model):
    def __init__(self, norm_par: dict[str, list[float]] = None,
                 num_classes: int = 1000, **kwargs):
        super().__init__(num_classes=num_classes, norm_par=norm_par, **kwargs)

    def define_preprocess(self, norm_par: dict[str, list[float]] = None, **kwargs):
        self.normalize = Normalize(mean=norm_par['mean'], std=norm_par['std']) \
            if norm_par is not None else nn.Identity()

    # get feature map
    # input: (batch_size, channels, height, width)
    # output: (batch_size, [feature_map])
    def get_fm(self, x: torch.Tensor) -> torch.Tensor:
        return self.features(self.normalize(x))


class ImageModel(Model):

    @classmethod
    def add_argument(cls, group: argparse._ArgumentGroup):
        super().add_argument(group)
        group.add_argument('--adv_train', choices=['pgd', 'free', 'trades'], help='adversarial training.')
        group.add_argument('--adv_train_random_init', action='store_true')
        group.add_argument('--adv_train_iter', type=int, help='adversarial training PGD iteration, defaults to 7.')
        group.add_argument('--adv_train_alpha', type=float, help='adversarial training PGD alpha, defaults to 2/255.')
        group.add_argument('--adv_train_eps', type=float, help='adversarial training PGD eps, defaults to 8/255.')
        group.add_argument('--adv_train_eval_iter', type=int)
        group.add_argument('--adv_train_eval_alpha', type=float)
        group.add_argument('--adv_train_eval_eps', type=float)
        group.add_argument('--adv_train_trades_beta', type=float,
                           help='regularization, i.e., 1/lambda in TRADES')

        group.add_argument('--norm_layer', choices=['bn', 'gn'], default='bn')
        group.add_argument('--sgm', action='store_true', help='whether to use sgm gradient, defaults to False')
        group.add_argument('--sgm_gamma', type=float, help='sgm gamma, defaults to 1.0')
        return group

    def __init__(self, name: str = 'imagemodel', layer: int = None,
                 model: Union[type[_ImageModel], _ImageModel] = _ImageModel, dataset: ImageSet = None, data_shape: list[int] = None,
                 adv_train: str = None, adv_train_random_init: bool = False, adv_train_eval_random_init: bool = None,
                 adv_train_iter: int = 7, adv_train_alpha: float = 2 / 255, adv_train_eps: float = 8 / 255,
                 adv_train_eval_iter: int = None, adv_train_eval_alpha: float = None, adv_train_eval_eps: float = None,
                 adv_train_trades_beta: float = 6.0,
                 norm_layer: str = 'bn', sgm: bool = False, sgm_gamma: float = 1.0,
                 norm_par: dict[str, list[float]] = None, suffix: str = None, **kwargs):
        name = self.get_name(name, layer=layer)
        if norm_par is None and isinstance(dataset, ImageSet):
            norm_par = None if dataset.normalize else dataset.norm_par
        if 'num_classes' not in kwargs.keys() and dataset is None:
            kwargs['num_classes'] = 1000
        if adv_train is not None and suffix is None:
            suffix = '_adv_train'
        super().__init__(name=name, model=model, dataset=dataset, data_shape=data_shape,
                         norm_par=norm_par, suffix=suffix, **kwargs)
        assert norm_layer in ['bn', 'gn']
        if norm_layer == 'gn':
            replace_bn_to_gn(self._model)

        if data_shape is None:
            assert isinstance(dataset, ImageSet)
            data_shape = dataset.data_shape
        args = {'padding': 3} if 'vgg' in name else {}  # TODO: so ugly
        set_first_layer_channel(self._model.features, channel=data_shape[0], **args)

        self.sgm: bool = sgm
        self.sgm_gamma: float = sgm_gamma
        self.adv_train = adv_train
        self.adv_train_random_init = adv_train_random_init
        self.adv_train_eval_random_init = adv_train_eval_random_init if adv_train_eval_random_init is not None else adv_train_random_init
        self.adv_train_iter = adv_train_iter
        self.adv_train_alpha = adv_train_alpha
        self.adv_train_eps = adv_train_eps
        self.adv_train_eval_iter = adv_train_eval_iter if adv_train_eval_iter is not None else adv_train_iter
        self.adv_train_eval_alpha = adv_train_eval_alpha if adv_train_eval_alpha is not None else adv_train_alpha
        self.adv_train_eval_eps = adv_train_eval_eps if adv_train_eval_eps is not None else adv_train_eps
        self.adv_train_trades_beta = adv_train_trades_beta

        self.param_list['imagemodel'] = []
        if sgm:
            self.param_list['imagemodel'].append('sgm_gamma')
        if adv_train is not None:
            if 'suffix' not in self.param_list['model']:
                self.param_list['model'].append('suffix')
            self.param_list['adv_train'] = ['adv_train', 'adv_train_random_init', 'adv_train_eval_random_init',
                                            'adv_train_iter', 'adv_train_alpha', 'adv_train_eps',
                                            'adv_train_eval_iter', 'adv_train_eval_alpha', 'adv_train_eval_eps']
            if adv_train == 'trades':
                self.param_list['adv_train'].append('adv_train_trades_beta')
            clip_min, clip_max = 0.0, 1.0
            if norm_par is None and isinstance(dataset, ImageSet):
                if dataset.normalize and dataset.norm_par is not None:
                    mean = torch.tensor(dataset.norm_par['mean'], device=env['device']).view(-1, 1, 1)
                    std = torch.tensor(dataset.norm_par['std'], device=env['device']).view(-1, 1, 1)
                    clip_min, clip_max = -mean / std, (1 - mean) / std
                    self.adv_train_eval_alpha /= std
                    self.adv_train_eval_eps /= std
                    self.adv_train_alpha /= std
                    self.adv_train_eps /= std
            self.pgd = PGD(pgd_alpha=self.adv_train_eval_alpha, pgd_eps=self.adv_train_eval_eps,
                           iteration=self.adv_train_eval_iter, stop_threshold=None,
                           target_idx=0,
                           random_init=self.adv_train_eval_random_init,
                           clip_min=clip_min, clip_max=clip_max,
                           model=self, dataset=self.dataset)
        self._model: _ImageModel
        self.dataset: ImageSet

    def trades_loss_fn(self, _input: torch.Tensor, org_prob: torch.Tensor, **kwargs):
        return -criterion_kl(log_softmax(self(_input)), org_prob)

    @classmethod
    def get_name(cls, name: str, layer: int = None) -> str:
        full_list = name.split('_')
        partial_name = full_list[0]
        re_list = re.findall(r'\d+|\D+', partial_name)
        if len(re_list) > 1:
            layer = int(re_list[1])
        elif layer is not None:
            partial_name += str(layer)
        full_list[0] = partial_name
        return '_'.join(full_list)

    # TODO: requires _input shape (N, C, H, W)
    # Reference: https://keras.io/examples/vision/grad_cam/
    def get_heatmap(self, _input: torch.Tensor, _label: torch.Tensor, method: str = 'grad_cam', cmap: Colormap = jet) -> torch.Tensor:
        squeeze_flag = False
        if _input.dim() == 3:
            _input = _input.unsqueeze(0)    # (N, C, H, W)
            squeeze_flag = True
        if isinstance(_label, int):
            _label = [_label] * len(_input)
        _label = torch.as_tensor(_label, device=_input.device)
        heatmap = _input    # linting purpose
        if method == 'grad_cam':
            feats = self._model.get_fm(_input).detach()   # (N, C', H', W')
            feats.requires_grad_()
            _output: torch.Tensor = self._model.pool(feats)   # (N, C', 1, 1)
            _output = self._model.flatten(_output)   # (N, C')
            _output = self._model.classifier(_output)   # (N, num_classes)
            _output = _output.gather(dim=1, index=_label.unsqueeze(1)).sum()
            grad = torch.autograd.grad(_output, feats)[0]   # (N, C',H', W')
            feats.requires_grad_(False)
            weights = grad.mean(dim=-2, keepdim=True).mean(dim=-1, keepdim=True)    # (N, C',1,1)
            heatmap = (feats * weights).sum(dim=1, keepdim=True).clamp(0)  # (N, 1, H', W')
            # heatmap.sub_(heatmap.min(dim=-2, keepdim=True)[0].min(dim=-1, keepdim=True)[0])
            heatmap.div_(heatmap.max(dim=-2, keepdim=True)[0].max(dim=-1, keepdim=True)[0])
            heatmap: torch.Tensor = F.upsample(heatmap, _input.shape[-2:], mode='bilinear')[:, 0]   # (N, H, W)
            # Note that we violate the image order convension (W, H, C)
        elif method == 'saliency_map':
            _input.requires_grad_()
            _output = self(_input).gather(dim=1, index=_label.unsqueeze(1)).sum()
            grad = torch.autograd.grad(_output, _input)[0]   # (N,C,H,W)
            _input.requires_grad_(False)

            heatmap = grad.abs().max(dim=1)[0]   # (N,H,W)
            heatmap.sub_(heatmap.min(dim=-2, keepdim=True)[0].min(dim=-1, keepdim=True)[0])
            heatmap.div_(heatmap.max(dim=-2, keepdim=True)[0].max(dim=-1, keepdim=True)[0])
        heatmap = apply_cmap(heatmap.detach().cpu(), cmap)
        return heatmap[0] if squeeze_flag else heatmap

    def get_data(self, data: tuple[torch.Tensor, torch.Tensor], adv_train: bool = False, **kwargs) -> tuple[torch.Tensor, torch.Tensor]:
        """In trainining process, `adv_train` args will not be passed to `get_data`. So it's always `False`."""
        _input, _label = super().get_data(data, **kwargs)
        if adv_train:
            assert self.pgd is not None
            adv_x, _ = self.pgd.optimize(_input=_input, target=_label)
            return adv_x, _label
        return _input, _label

    def _validate(self, adv_train: bool = None, **kwargs) -> tuple[float, float]:
        adv_train = adv_train if adv_train is not None else bool(self.adv_train)
        if not adv_train:
            return super()._validate(**kwargs)
        _, clean_acc = super()._validate(print_prefix='Validate Clean', main_tag='valid clean',
                                         adv_train=False, **kwargs)
        _, adv_acc = super()._validate(print_prefix='Validate Adv', main_tag='valid adv',
                                       adv_train=True, **kwargs)
        return adv_acc, clean_acc + adv_acc

    def _train(self, epoch: int, optimizer: Optimizer, lr_scheduler: _LRScheduler = None,
               grad_clip: float = None, pre_conditioner: Union[KFAC, EKFAC] = None, adv_train: bool = None,
               print_prefix: str = 'Epoch', start_epoch: int = 0, resume: int = 0,
               validate_interval: int = 10, save: bool = False, amp: bool = False,
               loader_train: torch.utils.data.DataLoader = None, loader_valid: torch.utils.data.DataLoader = None,
               epoch_fn: Callable[..., None] = None,
               get_data_fn: Callable[..., tuple[torch.Tensor, torch.Tensor]] = None,
               loss_fn: Callable[..., torch.Tensor] = None,
               after_loss_fn: Callable[..., None] = None,
               validate_fn: Callable[..., tuple[float, float]] = None,
               save_fn: Callable[..., None] = None, file_path: str = None, folder_path: str = None, suffix: str = None,
               writer=None, main_tag: str = 'train', tag: str = '',
               accuracy_fn: Callable[..., list[float]] = None,
               verbose: bool = True, indent: int = 0, **kwargs):
        adv_train = adv_train if adv_train is not None else bool(self.adv_train)
        if adv_train:
            after_loss_fn_old = after_loss_fn
            if not callable(after_loss_fn) and hasattr(self, 'after_loss_fn'):
                after_loss_fn_old = getattr(self, 'after_loss_fn')
            loss_fn = loss_fn if callable(loss_fn) else self.loss

            def after_loss_fn_new(_input: torch.Tensor, _label: torch.Tensor, _output: torch.Tensor,
                                  loss: torch.Tensor, optimizer: Optimizer, loss_fn: Callable[..., torch.Tensor] = None,
                                  amp: bool = False, scaler: torch.cuda.amp.GradScaler = None, **kwargs):
                optimizer.zero_grad()
                self.zero_grad()
                if pre_conditioner is not None:
                    pre_conditioner.reset()

                if self.adv_train == 'free':
                    noise = self.pgd.init_noise(_input.shape, pgd_eps=self.adv_train_eps,
                                                random_init=self.adv_train_random_init,
                                                device=_input.device)
                    adv_x = add_noise(_input=_input, noise=noise, batch=self.pgd.universal,
                                      clip_min=self.pgd.clip_min, clip_max=self.pgd.clip_max)
                    noise.data = self.pgd.valid_noise(adv_x, _input)
                    for m in range(self.adv_train_iter):
                        loss = loss_fn(adv_x, _label)
                        if amp:
                            scaler.scale(loss).backward()
                            scaler.step(optimizer)
                            scaler.update()
                        else:
                            loss.backward()
                            optimizer.step()
                        optimizer.zero_grad()
                        self.zero_grad()
                        # self.eval()
                        adv_x, _ = self.pgd.optimize(_input=_input, noise=noise, target=_label, iteration=1,
                                                     pgd_alpha=self.adv_train_alpha, pgd_eps=self.adv_train_eps)
                        # self.train()
                        loss = loss_fn(adv_x, _label)
                else:
                    loss = self.adv_loss(_input=_input, _label=_label, loss_fn=loss_fn)

                if amp:
                    scaler.scale(loss).backward()
                else:
                    loss.backward()
                if callable(after_loss_fn_old):
                    after_loss_fn_old(_input=_input, _label=_label, _output=_output,
                                      loss=loss, optimizer=optimizer, loss_fn=loss_fn,
                                      amp=amp, scaler=scaler, **kwargs)
            after_loss_fn = after_loss_fn_new

        super()._train(epoch=epoch, optimizer=optimizer, lr_scheduler=lr_scheduler,
                       grad_clip=grad_clip, pre_conditioner=pre_conditioner, adv_train=adv_train,
                       print_prefix=print_prefix, start_epoch=start_epoch, resume=resume,
                       validate_interval=validate_interval, save=save, amp=amp,
                       loader_train=loader_train, loader_valid=loader_valid,
                       epoch_fn=epoch_fn, get_data_fn=get_data_fn, loss_fn=loss_fn, after_loss_fn=after_loss_fn, validate_fn=validate_fn,
                       save_fn=save_fn, file_path=file_path, folder_path=folder_path, suffix=suffix,
                       writer=writer, main_tag=main_tag, tag=tag,
                       accuracy_fn=accuracy_fn, verbose=verbose, indent=indent, **kwargs)

    def adv_loss(self, _input: torch.Tensor, _label: torch.Tensor,
                 loss_fn: Callable[..., torch.Tensor] = None, adv_train: str = None) -> torch.Tensor:
        adv_train = adv_train if adv_train is not None else self.adv_train
        loss_fn = loss_fn if callable(loss_fn) else self.loss
        if adv_train == 'trades':
            noise = 1e-3 * torch.randn_like(_input)
            org_prob = self.get_prob(_input)
            adv_x, _ = self.pgd.optimize(_input=_input, noise=noise, target=_label,
                                         iteration=self.adv_train_iter,
                                         pgd_alpha=self.adv_train_alpha,
                                         pgd_eps=self.adv_train_eps,
                                         loss_fn=self.trades_loss_fn,
                                         loss_kwargs={'org_prob': org_prob.detach()})
            adv_x = _input + (adv_x - _input).detach()
            return loss_fn(_input, _label) - self.adv_train_trades_beta * \
                self.trades_loss_fn(_input=adv_x, org_prob=org_prob)
        else:
            adv_x, _ = self.pgd.optimize(_input=_input, target=_label,
                                         iteration=self.adv_train_iter,
                                         pgd_alpha=self.adv_train_alpha,
                                         pgd_eps=self.adv_train_eps,
                                         random_init=self.adv_train_random_init)
            adv_x = _input + (adv_x - _input).detach()
            return loss_fn(adv_x, _label)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import autovul.vision.utils.model_archs.enas.mutables as mutables
from autovul.vision.utils.model_archs.enas.ops import FactorizedReduce, ConvBranch, PoolBranch

import torch.nn as nn


class ENASLayer(mutables.MutableScope):

    def __init__(self, key, prev_labels, in_filters, out_filters):
        super().__init__(key)
        self.in_filters = in_filters
        self.out_filters = out_filters
        self.mutable = mutables.LayerChoice([
            ConvBranch(in_filters, out_filters, 3, 1, 1, separable=False),
            ConvBranch(in_filters, out_filters, 3, 1, 1, separable=True),
            ConvBranch(in_filters, out_filters, 5, 1, 2, separable=False),
            ConvBranch(in_filters, out_filters, 5, 1, 2, separable=True),
            PoolBranch('avg', in_filters, out_filters, 3, 1, 1),
            PoolBranch('max', in_filters, out_filters, 3, 1, 1)
        ])
        if len(prev_labels) > 0:
            self.skipconnect = mutables.InputChoice(choose_from=prev_labels, n_chosen=None)
        else:
            self.skipconnect = None
        self.batch_norm = nn.BatchNorm2d(out_filters, affine=False)

    def forward(self, prev_layers):
        out = self.mutable(prev_layers[-1])
        if self.skipconnect is not None:
            connection = self.skipconnect(prev_layers[:-1])
            if connection is not None:
                out += connection
        return self.batch_norm(out)


class GeneralNetwork(nn.Module):
    def __init__(self, num_layers=12, out_filters=24, in_channels=3, num_classes=10,
                 dropout_rate=0.0):
        super().__init__()
        self.num_layers = num_layers
        self.num_classes = num_classes
        self.out_filters = out_filters

        self.stem = nn.Sequential(
            nn.Conv2d(in_channels, out_filters, 3, 1, 1, bias=False),
            nn.BatchNorm2d(out_filters)
        )

        pool_distance = self.num_layers // 3
        self.pool_layers_idx = [pool_distance - 1, 2 * pool_distance - 1]
        self.dropout_rate = dropout_rate
        self.dropout = nn.Dropout(self.dropout_rate)

        self.layers = nn.ModuleList()
        self.pool_layers = nn.ModuleList()
        labels = []
        for layer_id in range(self.num_layers):
            labels.append("layer_{}".format(layer_id))
            if layer_id in self.pool_layers_idx:
                self.pool_layers.append(FactorizedReduce(self.out_filters, self.out_filters))
            self.layers.append(ENASLayer(labels[-1], labels[:-1], self.out_filters, self.out_filters))

        self.gap = nn.AdaptiveAvgPool2d(1)
        self.flatten = nn.Flatten()
        self.dense = nn.Linear(self.out_filters, self.num_classes)

    def forward(self, x):
        cur = self.stem(x)
        layers = [cur]
        for layer_id in range(self.num_layers):
            cur = self.layers[layer_id](layers)
            layers.append(cur)
            if layer_id in self.pool_layers_idx:
                for i, layer in enumerate(layers):
                    layers[i] = self.pool_layers[self.pool_layers_idx.index(layer_id)](layer)
                cur = layers[-1]

        cur = self.gap(cur)
        cur = self.flatten(cur)
        cur = self.dropout(cur)
        logits = self.dense(cur)
        return logits
